#include "charAndSeqModel.h"
#include "kMean.h"
#include "someUtil.h"
#include "errorMsg.h"
#include "matrixUtils.h"
#include "likelihoodComputation.h"
#include "treeIt.h"
#include "treeUtil.h"
#include "gammaDistribution.h"
#include "hky.h"
#include "gtrModel.h"
#include "talRandom.h"
#include "evaluateCharacterFreq.h"

using namespace std;
#define lowerValueOfRateParam 0.01
#define upperValueOfRateParam 20.0
#define lowerValueOfDistrParam 0.1
#define upperValueOfDistrParam 10.0
#define lowerValueOfProbRateParam 0.01
#define upperValueOfProbRateParam 1.0
#define upperValueOfScaleParam 5.0
#define lowerValueOfScaleParam 0.2




/***********constructor***************/
charAndSeqModel::charAndSeqModel(stochasticProcess* pCharModel, stochasticProcess* pSeqModel, MDOUBLE relRate, seqModelType nucModel, combinedModelType modelType, MDOUBLE treeLength)
:_relRate(relRate), _nucModel(nucModel), _modelType(modelType), _param2Bound(1.0), _param1Bound(0.05), _treeLengthBound(1.0), _treeLength(treeLength)
{
	_pCharModel = pCharModel->clone();
	_pSeqModel = pSeqModel->clone();
}


charAndSeqModel::charAndSeqModel(const charAndSeqModel &other)
: _pCharModel(NULL),_pSeqModel(NULL) 
{
	copy(&other);
}

charAndSeqModel& charAndSeqModel::operator=(const charAndSeqModel &other)
{
	copy(&other);
	return *this;
}

/***********destructor***************/
charAndSeqModel::~charAndSeqModel()
{
	if (_pCharModel)
		delete _pCharModel;
	if (_pSeqModel)
		delete _pSeqModel;
}

void charAndSeqModel::copy(const charAndSeqModel* pOther)
{
	if (_pCharModel)
		delete _pCharModel;
	if (_pSeqModel)
		delete _pSeqModel;
	_pCharModel = NULL;
	_pSeqModel = NULL;

	if (pOther->_pCharModel)
		_pCharModel = pOther->_pCharModel->clone();
    if (pOther->_pSeqModel)
		_pSeqModel = pOther->_pSeqModel->clone();

	_relRate = pOther->_relRate; 
	_nucModel = pOther->_nucModel;
	_modelType = pOther->_modelType;
	_param2Bound = pOther->_param2Bound;
	_param1Bound = pOther->_param1Bound;
	_treeLengthBound = pOther->_treeLengthBound;
	_treeLength = pOther->_treeLength;
}


void charAndSeqModel::setRandomRateParams()
{
	vector<paramType> params;
	params.push_back(RELATIVE_RATE);
	getSeqModelParams(params);
	getCharModelParams(params);
	for (int i = 0; i < params.size(); ++i)
	{
		MDOUBLE val = talRandom::giveRandomNumberBetweenTwoPoints(getParamLowerBound(params[i]), getParamUpperBound(params[i]));
		setParam(params[i], val);
	}
}


//scale multiple trees according to the corresponding mapping
void charAndSeqModel::scaleTrees(const vector<tree>& trees, const vector<treeMutationHistory>& mappings, MDOUBLE relRate, vector<tree> &outtrees) const
{
	for (int t = 0; t < trees.size(); ++t)
	{
		vector<treeMutationHistory> oneMapping;
		vector<tree> oneTree;
		oneMapping.push_back(mappings[t]);
        scaleTrees(trees[t], oneMapping, relRate, oneTree);
		outtrees.push_back(oneTree[0]);
	}
}


void charAndSeqModel::scaleTrees(const tree& inTree, const vector<treeMutationHistory>& mappings, MDOUBLE relRate, vector<tree> &outtrees) const
{
	int mappingNum = mappings.size();
	for (int i = 0; i < mappingNum; ++i)
	{
		//create a new tree scaled according to the mappings of character 1
        tree relTree(inTree);
		treeIterTopDownConst tIt(relTree);
		for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
		{
			if (mynode->isRoot())
				continue;
			int nodeId = mynode->id();
			mutationHistoryPerBranch branchHistory = mappings[i][nodeId];
			MDOUBLE time0 = 0.0; //time spent in state 0
			MDOUBLE time1 = 0.0; //time spent in state 1
			for (int m = 0; m < branchHistory.size(); ++m)
			{
				if (branchHistory[m].first == 0)
					time0 += branchHistory[m].second;
				else
					time1 += branchHistory[m].second;
			}
			if (!DEQUAL(time0 + time1, mynode->dis2father()))
                errorMsg::reportError("total mapping on branch not equal to branch length");
			MDOUBLE prop1 = time1/(time0 + time1);
			MDOUBLE prop0 = 1.0 - prop1;
			MDOUBLE newLen = mynode->dis2father() * (prop1 * relRate + prop0);
			mynode->setDisToFather(newLen);
		}

		//keep all scaled trees to have the same total tree length
		MDOUBLE curTreeLength = getSumOfBranchLengths(relTree);
		if (curTreeLength != _treeLength)
            relTree.multipleAllBranchesByFactor(_treeLength /curTreeLength);
		
		outtrees.push_back(relTree);
	}
}



//1. generates mappingNum histories (stochastic character mappings)
//2. Based on each history - generate a tree by scaling each branch by p1, the proportion of time it was in character 1, times the relative rte parameter _relRate
//		new_length = OldLength*p1*_relRate + OldLength*(1-p1)
//3. calculate likelihood based on the set of generated trees
MDOUBLE charAndSeqModel::compLogLikelihood(const tree& inTree, const sequenceContainer& charContainer, const sequenceContainer& seqContainer, int mappingNum, int heuristicGroups, const vector<tree> &charTrees, llCalc llType)
{
	//if (_modelType == M0)
	//	return compLogLikelihood(inTree, seqContainer, charTrees, charContainer, llType);
	stochasticMapping sm(inTree, *_pCharModel);
	vector<treeMutationHistory> mappings;
	sm.generateStochasticMapping(charContainer, 0, mappingNum, mappings);
	return compLogLikelihood(inTree, mappings, seqContainer, heuristicGroups, charTrees, charContainer, llType);
}

//computes the combined model likelihood, which involves stochastic mapping, based on a vector of possible trees
MDOUBLE charAndSeqModel::compLogLikelihood(const vector<tree> &trees, const sequenceContainer& charContainer, const sequenceContainer& seqContainer, int mappingNum, int heuristicGroups, const vector<tree> &charTrees, llCalc llType)
{
	int treeNum = trees.size();
	if (treeNum == 1)
		return compLogLikelihood(trees[0], charContainer, seqContainer, mappingNum, heuristicGroups, charTrees, llType);
    else if (treeNum != mappingNum)
		errorMsg::reportError("cannot compute likelihood: assuming the vector of trees is equal to the number of mappings");
	if (_modelType == M0)
		return compLogLikelihood(trees, seqContainer, charTrees, charContainer, llType);
	//if (heuristicGroups != 0)
	//	errorMsg::reportError("cannot compute heuristic likelihood for multiple trees");

	vector<treeMutationHistory> allMappings;
	for (int t = 0; t < mappingNum; ++t)
	{
		stochasticMapping sm(trees[t], *_pCharModel);
		vector<treeMutationHistory> oneMappng;
		sm.generateStochasticMapping(charContainer, 0, 1, oneMappng);
		allMappings.push_back(oneMappng[0]);
	}
	return compLogLikelihood(trees, allMappings, seqContainer, heuristicGroups, charTrees, charContainer, llType);
}


//computes the likelihood of the sequence model based on a vector of stochastic mappings - one for each tree in the input trees
MDOUBLE charAndSeqModel::compLogLikelihood(const vector<tree> &trees, const vector<treeMutationHistory>& allMappings, const sequenceContainer &seqContainer, int heuristicGroups, const vector<tree> &CharTrees, const sequenceContainer &charContainer, llCalc llType)
{
	vector<tree> scaledTrees;
	scaleTrees(trees, allMappings, _relRate, scaledTrees);
	return compLogLikelihood(scaledTrees, seqContainer, heuristicGroups, CharTrees, charContainer, llType);
}


//computes the likelihood of the sequence model based on a vector of stochastic mappings
MDOUBLE charAndSeqModel::compLogLikelihood(const tree& inTree, const vector<treeMutationHistory>& mappings, const sequenceContainer &seqContainer, int heuristicGroups, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType)
{
	vector<tree> scaledTrees;
	scaleTrees(inTree, mappings, _relRate, scaledTrees);
	return compLogLikelihood(scaledTrees, seqContainer, heuristicGroups, charTrees, charContainer, llType);
}

//computes the likelihood of the sequence model based on a vector of trees that are the result of the stochastic mapping, and according to some heurisitc criterion. 
MDOUBLE charAndSeqModel::compLogLikelihood(const vector<tree> &trees, const sequenceContainer &seqContainer, int heuristicGroups, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType)
{
	int treesNum = trees.size();
	if (treesNum < 1)
		errorMsg::reportError("cannot compute likelihood without any tree");
	if (heuristicGroups == 1)
		return compLogLikelihoodAvgTree(trees, seqContainer, charTrees, charContainer, llType);
	else if (heuristicGroups > 1)
	{
		Vdouble weights;
		vector<tree> centerTrees;
		getKMeanTrees(trees, heuristicGroups, centerTrees, weights);
		return compLogLikelihood(centerTrees, seqContainer, weights, charTrees, charContainer, llType);
	}
	else return compLogLikelihood(trees, seqContainer, charTrees, charContainer, llType);
}


//this method gets a vector of trees - each as the result of the stochastic mapping scaling of the branch lengths.
//The resulting likelihood is the average over the likelihood of all trees
MDOUBLE charAndSeqModel::compLogLikelihood(const vector<tree> &seqTrees, const sequenceContainer &seqContainer, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType)
{
	//for debug
	twoStateModel* pCharModel = static_cast<twoStateModel*>(_pCharModel->getPijAccelerator()->getReplacementModel());
	MDOUBLE alpha = static_cast<gammaDistribution*>(_pSeqModel->distr())->getAlpha();
	hky* pSeqModel = static_cast<hky*>(_pSeqModel->getPijAccelerator()->getReplacementModel());
	MDOUBLE kappa = pSeqModel->getTrTv();
	MDOUBLE pi1 = pCharModel->getPi1();
	MDOUBLE rc = pCharModel->getScaleRate();
	MDOUBLE rr = getRelativeRate();

	///
	if (seqTrees.size() != charTrees.size())
		errorMsg::reportError("cannot compute likelihood: number of species trees and sequence trees are not equal");
	doubleRep sum = 0.0;
	for (int i = 0; i < seqTrees.size(); ++i)
	{
		doubleRep l = getTreeLikelihoodAllPos(seqTrees[i], seqContainer, *_pSeqModel);
		LOG(11, <<"LL sequence= "<<log(l)<<" al = "<<alpha<<" kap = "<<kappa<<" rr = "<<rr<<endl;); 
		if (llType == COMBINED) {
			doubleRep charL = getTreeLikelihoodAllPos(charTrees[i], charContainer, *_pCharModel);
			LOG(11, <<"LL char= "<<log(charL)<<" pi = "<<pi1<<" rc = "<<rc<<endl;); 
			l *= charL;
		}

		sum += l;
	}
	MDOUBLE avgL = log(sum / seqTrees.size());
	return avgL;
}

//computes the likelihood of the sequence model only based on a vector of trees that are the result of the stochastic mapping AND weight for each tree
MDOUBLE charAndSeqModel::compLogLikelihood(const vector<tree> &trees, const sequenceContainer &seqContainer, const Vdouble& weights, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType)
{
	//for debug
	twoStateModel* pCharModel = static_cast<twoStateModel*>(_pCharModel->getPijAccelerator()->getReplacementModel());
	MDOUBLE alpha = static_cast<gammaDistribution*>(_pSeqModel->distr())->getAlpha();
	hky* pSeqModel = static_cast<hky*>(_pSeqModel->getPijAccelerator()->getReplacementModel());
	MDOUBLE kappa = pSeqModel->getTrTv();
	MDOUBLE pi1 = pCharModel->getPi1();
	MDOUBLE rc = pCharModel->getScaleRate();
	MDOUBLE rr = getRelativeRate();

	///
	if (weights.size() != trees.size())
		errorMsg::reportError("the weights and trees vectors should be same size");
	
	doubleRep sum = 0.0;
	MDOUBLE totalWeights = 0.0;
	for (int i = 0; i < trees.size(); ++i)
	{
		doubleRep l = getTreeLikelihoodAllPos(trees[i], seqContainer, *_pSeqModel);
		LOG(11, <<"LL sequence= "<<log(l)<<" al = "<<alpha<<" kap = "<<kappa<<" rr = "<<rr<<endl;);  
		if (llType == COMBINED) {
			doubleRep charL = getTreeLikelihoodAllPos(charTrees[i], charContainer, *_pCharModel);
			LOG(11, <<"LL char= "<<log(charL)<<" pi = "<<pi1<<" rc = "<<rc<<endl;); 
			l *= charL;
		}
		sum += l * weights[i];
		totalWeights += weights[i];
	}
	MDOUBLE avgL = log(sum / totalWeights);
	return avgL;
}


//heuristic LL computation: get the average scaled tree from all mappings and computes LL based on the averaged scaled tree
MDOUBLE charAndSeqModel::compLogLikelihoodAvgTree(const vector<tree>& scaledTrees, const sequenceContainer& seqContainer, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType)
{
	tree avgTree = getAvgScaledTree(scaledTrees);
	vector<tree> oneTree;
	oneTree.push_back(avgTree);
	vector<tree> oneCharTree;
	oneCharTree.push_back(charTrees[0]);
	return compLogLikelihood(oneTree, seqContainer, oneCharTree, charContainer, llType);
}

doubleRep charAndSeqModel::getTreeLikelihoodAllPos(const tree& et, const sequenceContainer& sc,
																  const stochasticProcess& sp,
																  const Vdouble * const weights)
{
	computePijGam pi;
	pi.fillPij(et,sp);
	doubleRep res =1.0;
	doubleRep LofPos;
	int k;
	for (k=0; k < sc.seqLen(); ++k) {
		LofPos = likelihoodComputation::getLofPos(k,//pos,
			et,		//const tree& 
			sc,		// sequenceContainer& sc,
			pi,		//const computePijGam& ,
			sp
			/*unObservableData_p*/);
		res *= LofPos;//const stochasticProcess& );
	}

	return res;
}



MDOUBLE charAndSeqModel::compLogLikelihood(const tree& inTree, const sequenceContainer& seqContainer, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType)
{
	vector<tree> tmpVec;
	tmpVec.push_back(inTree);
	return compLogLikelihood(tmpVec, seqContainer, charTrees, charContainer, llType);
}


MDOUBLE charAndSeqModel::getParam(paramType type) const
{
	MDOUBLE res = -1;
	switch (type)
	{
		case TREE_LENGTH:
			res = getTreeLength();
			break;
		//case GLOBAL_RATE:
		//	res = getGlobalRate();
		//	break;
		case RELATIVE_RATE:
			res = getRelativeRate();
			break;
		case CHAR_MODEL_PARAM1:
			res = static_cast<twoStateModel*>(_pCharModel->getPijAccelerator()->getReplacementModel())->getPi1();
			break;
		case CHAR_MODEL_PARAM2:
			res = static_cast<twoStateModel*>(_pCharModel->getPijAccelerator()->getReplacementModel())->getScaleRate();
			break;
		case SEQ_MODEL_ASRV:
			res = static_cast<gammaDistribution*>(_pSeqModel->distr())->getAlpha();
			break;
		case HKY_MODEL_PARAM:
			res = static_cast<hky*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->getTrTv();
			break;
		case GTR_MODEL_PARAM1:
			res = static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->get_a2c();
			break;
		case GTR_MODEL_PARAM2:
			res = static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->get_a2g();
			break;
		case GTR_MODEL_PARAM3:
			res = static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->get_a2t();
			break;
		case GTR_MODEL_PARAM4:
			res = static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->get_c2g();
			break;
		case GTR_MODEL_PARAM5:
			res = static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->get_c2t();
			break;
		case GTR_MODEL_PARAM6:
			res = static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->get_g2t();
			break;
		default: 
			errorMsg::reportError("error in charAndSeqModel::getParam unknown param type");
			break;
	}
	return res;
}

void charAndSeqModel::setParam(paramType type, MDOUBLE val)
{
	switch (type)
	{
		case TREE_LENGTH:
			setTreeLength(val);
			break;
		//case GLOBAL_RATE:
		//	setGlobalRate(val);
		//	break;
		case RELATIVE_RATE:
			setRelativeRate(val);
			break;
		case CHAR_MODEL_PARAM1:
			static_cast<twoStateModel*>(_pCharModel->getPijAccelerator()->getReplacementModel())->setPi1(val);
			break;
		case CHAR_MODEL_PARAM2:
			static_cast<twoStateModel*>(_pCharModel->getPijAccelerator()->getReplacementModel())->setScaleRate(val);
			break;
		case SEQ_MODEL_ASRV:
			static_cast<gammaDistribution*>(_pSeqModel->distr())->setAlpha(val);
			break;
		case HKY_MODEL_PARAM:
			static_cast<hky*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->changeTrTv(val);
			break;
		case GTR_MODEL_PARAM1:
			static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->set_a2c(val);
			break;
		case GTR_MODEL_PARAM2:
			static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->set_a2g(val);
			break;
		case GTR_MODEL_PARAM3:
			static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->set_a2t(val);
			break;
		case GTR_MODEL_PARAM4:
			static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->set_c2g(val);
			break;
		case GTR_MODEL_PARAM5:
			static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->set_c2t(val);
			break;
		case GTR_MODEL_PARAM6:
			static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel())->set_g2t(val);
			break;
		default: 
			errorMsg::reportError("error in charAndSeqModel::setParam unknown param type");
			break;
	}
}

string charAndSeqModel::getParamName(paramType type) const
{
	string res("");
	switch (type)
	{
		case TREE_LENGTH:
			res = "TREE_LENGTH";
			break;
		case GLOBAL_RATE:
			res = "GLOBAL_RATE";
			break;
		case RELATIVE_RATE:
			res = "RELATIVE_RATE";
			break;
		case CHAR_MODEL_PARAM1:
			res = "CHAR_MODEL_PARAM1";
			break;
		case CHAR_MODEL_PARAM2:
			res = "CHAR_MODEL_PARAM2";
			break;
		case SEQ_MODEL_ASRV:
			res = "SEQ_MODEL_ASRV";
			break;
		case HKY_MODEL_PARAM:
			res = "HKY_MODEL_PARAM";
			break;
		case GTR_MODEL_PARAM1:
			res = "GTR_MODEL_PARAM1";
			break;
		case GTR_MODEL_PARAM2:
			res = "GTR_MODEL_PARAM2";
			break;
		case GTR_MODEL_PARAM3:
			res = "GTR_MODEL_PARAM3";
			break;
		case GTR_MODEL_PARAM4:
			res = "GTR_MODEL_PARAM4";
			break;
		case GTR_MODEL_PARAM5:
			res = "GTR_MODEL_PARAM5";
			break;
		case GTR_MODEL_PARAM6:
			res = "GTR_MODEL_PARAM6";
			break;

		default: 
			errorMsg::reportError("error in charAndSeqModel::getParamName unknown param type");
			break;
	}
	return res;
}

MDOUBLE charAndSeqModel::getParamLowerBound(paramType type) const
{
	MDOUBLE res = -1.0;
	switch (type)
	{
	case GLOBAL_RATE:
	case RELATIVE_RATE:
	case HKY_MODEL_PARAM:
	case GTR_MODEL_PARAM1:
	case GTR_MODEL_PARAM2:
	case GTR_MODEL_PARAM3:
	case GTR_MODEL_PARAM4:
	case GTR_MODEL_PARAM5:
	case GTR_MODEL_PARAM6:
		res = lowerValueOfRateParam;
		break;
	case SEQ_MODEL_ASRV:
		res = lowerValueOfDistrParam;
		break;
	case CHAR_MODEL_PARAM1:
        //res = lowerValueOfProbRateParam;
		res = _param1Bound;  
		break;
	case CHAR_MODEL_PARAM2:
		res = lowerValueOfScaleParam * _param2Bound;
		break;
	case TREE_LENGTH:
		res = lowerValueOfScaleParam * _treeLengthBound;
		break;
	}
	return res;
}

MDOUBLE charAndSeqModel::getParamUpperBound(paramType type) const
{
	MDOUBLE res = -1.0;
	switch (type)
	{
	case GLOBAL_RATE:
	case RELATIVE_RATE:
	case HKY_MODEL_PARAM:
	case GTR_MODEL_PARAM1:
	case GTR_MODEL_PARAM2:
	case GTR_MODEL_PARAM3:
	case GTR_MODEL_PARAM4:
	case GTR_MODEL_PARAM5:
	case GTR_MODEL_PARAM6:
		res = upperValueOfRateParam;
		break;
	case SEQ_MODEL_ASRV:
		res = upperValueOfDistrParam;
		break;
	case CHAR_MODEL_PARAM1:
		//res = upperValueOfProbRateParam;
		res = 1.0 - _param1Bound; 
		break;
	case CHAR_MODEL_PARAM2:
		res = upperValueOfScaleParam * _param2Bound;
		break;
	case TREE_LENGTH:
		res = upperValueOfScaleParam * _treeLengthBound;
		break;
	}
	return res;
}

void charAndSeqModel::printModelParams(ostream &out) const
{
	twoStateModel* pCharModel = static_cast<twoStateModel*>(_pCharModel->getPijAccelerator()->getReplacementModel());

	out<<"#model params:"<<endl;
	if (_modelType != M0)
	{
        out <<"charModelParam1="<<pCharModel->getPi1()<<"\t";
		out <<"charModelParam2="<<pCharModel->getScaleRate()<<"\t";
		out <<"relativeRate="<<getRelativeRate()<<"\t";
	}
	out <<"treeLength="<<getTreeLength()<<"\t";
	if (_pSeqModel->categories() > 1)
		out <<"alpha="<<static_cast<gammaDistribution*>(_pSeqModel->distr())->getAlpha()<<"\t";
	switch (_nucModel)
	{
		case JC:
            break;
		case K2:
		case HKY:
			{
                hky* pSeqModel = static_cast<hky*>(_pSeqModel->getPijAccelerator()->getReplacementModel());
				out <<"kappa="<<pSeqModel->getTrTv()<<"\t";
			}
			break;
		case GTR:
			{
				gtrModel* pSeqModel = static_cast<gtrModel*>(_pSeqModel->getPijAccelerator()->getReplacementModel());
				out <<"a2c="<<pSeqModel->get_a2c()<<"\t";
				out <<"a2g="<<pSeqModel->get_a2g()<<"\t";
				out <<"a2t="<<pSeqModel->get_a2t()<<"\t";
				out <<"c2g="<<pSeqModel->get_c2g()<<"\t";
				out <<"c2t="<<pSeqModel->get_c2t()<<"\t";
				out <<"g2t="<<pSeqModel->get_g2t()<<"\t";
			}
			break;
	}
	out<<endl;
}


void charAndSeqModel::getSeqModelParams(vector<paramType>& params)
{
	if (getSeqProcess()->categories() > 1)
        params.push_back(SEQ_MODEL_ASRV);

	switch (_nucModel)
	{
	case JC:
		break;
		case K2:
		case HKY:
			params.push_back(HKY_MODEL_PARAM);
			break;
		case GTR:
			params.push_back(GTR_MODEL_PARAM1);
			params.push_back(GTR_MODEL_PARAM2);
			params.push_back(GTR_MODEL_PARAM3);
			params.push_back(GTR_MODEL_PARAM4);
			params.push_back(GTR_MODEL_PARAM5);
			params.push_back(GTR_MODEL_PARAM6);
			break;
		case JTT:
		case MT_REV:
		case CP_REV:
		case LG:
		case WAG:
		case AA_JC:
			break;
	}
}

void charAndSeqModel::getCharModelParams(vector<paramType>& params)
{
	params.push_back(CHAR_MODEL_PARAM1);
	params.push_back(CHAR_MODEL_PARAM2);
}


tree charAndSeqModel::getAvgScaledTree(const vector<tree>& scaledTrees)
{
	if (_modelType == M0)
		errorMsg::reportError("not supposed to be here");

	//go over all scaled trees and for each branch calculate sum over all mappings
	Vdouble branchesSum(scaledTrees[0].getNodesNum(), 0.0);

	for (int i = 0; i < scaledTrees.size(); ++i)
	{
		//create a new tree scaled according to the mappings of character 1
		treeIterTopDownConst tIt(scaledTrees[i]);
		for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
		{
			if (mynode->isRoot())
				continue;
			int nodeId = mynode->id();
			MDOUBLE dis = mynode->dis2father();
			branchesSum[nodeId] += dis;
		}
	}

	tree relTree(scaledTrees[0]);
	treeIterTopDownConst tIt2(relTree);
	for (tree::nodeP mynode = tIt2.first(); mynode != tIt2.end(); mynode = tIt2.next()) 
	{
		if (mynode->isRoot())
			continue;
		int nodeId = mynode->id();
		MDOUBLE avgDis = branchesSum[nodeId] / scaledTrees.size();
		mynode->setDisToFather(avgDis);
	}
	return relTree;
}


tree charAndSeqModel::getAvgScaledTree(const tree& inTree, const sequenceContainer& charContainer, int mappingNum)
{
	if (_modelType == M0)
		return inTree;

	stochasticMapping sm(inTree, *_pCharModel);
	vector<treeMutationHistory> mappings;
	sm.generateStochasticMapping(charContainer, 0, mappingNum, mappings);

	vector<tree> scaledTrees;
	scaleTrees(inTree, mappings, _relRate, scaledTrees);

	return getAvgScaledTree(scaledTrees);

}


void charAndSeqModel::getKMeanTrees(const vector<tree> &trees,  int k, vector<tree>& centerTrees, Vdouble& weights)
{
	k = k-1;
	int iterNum = 2;
	int treesNum = trees.size();
	//for each scaled tree - create a vector of branch lengths
	VVdouble points;
	int nodesNum = trees[0].getNodesNum();
	resizeMatrix(points, treesNum, nodesNum);
	vector <tree::nodeP> nodes;
	int t = 0;
	for (; t < treesNum; ++t) 
	{
		trees[t].getAllNodes(nodes, trees[t].getRoot());
		for (int b = 0; b < nodesNum; ++b)
			points[t][b] = nodes[b]->dis2father();
	}
			
	//send to k-clustering
	kMean cluster;
	cluster.run(points, k, iterNum);
	const Vint centers = cluster.getCenters();
	const VVint assignments = cluster.getClusterAssignments();
	//organize the new trees and weights;
	weights.resize(k);
	centerTrees.clear();
	for (t = 0; t < k; ++t)
	{
		centerTrees.push_back(trees[centers[t]]);
		weights[t] = assignments[t].size();
	}
}

//set upper/lower bound for character model param 1.
//The bound is set so that the frequency of either charaqcter won't be too much lower than the frequency of the less common character
void charAndSeqModel::setParam1Bound (const sequenceContainer& charContainer)
{
	VVint counts4pos;
	getCharacterCounts(charContainer, counts4pos);
	MDOUBLE lowerFreq = 10.0;
	MDOUBLE taxaNum = static_cast<MDOUBLE>(charContainer.numberOfSeqs());
	for (int c = 0; c < charContainer.alphabetSize(); ++c)
	{
		int count = counts4pos[0][c];
		MDOUBLE freq = static_cast<MDOUBLE>(count) / taxaNum;
		if (freq < lowerFreq)
			lowerFreq = freq;
	}
	
	_param1Bound = max(1.0 / taxaNum, lowerFreq/4.0);
}
