#ifndef __CHAR_SEQ_MODEL
#define __CHAR_SEQ_MODEL
#include "traitRateOptions.h" //only for seqModeltype
#include "definitions.h"
#include "tree.h"
#include "sequenceContainer.h"
#include "stochasticProcess.h"
#include "stochasticMapping.h"


/***********************************************************************
This class represents the combined binary-character + molecular sequenc process.
For character: the is a two state model
For sequence: the model can be any nucleotide model. Currently we allow for JC, KIMURA 2 param, HKY and GTR. There is another component of rate variation: gamma or homogenous.

The likelihood calculations are as follows:
1. Stochastic mapping simulations are used to map the identity of the binary characters on the tree
2. All segments of the tree for which character 1 was simulated are multiplied by _relRate (the rate of character 1 relative to character 0)
3. For one mapping: the likelihood is calculated based on the molecular model and the tree (that is the result of the character model).
4. The total likelihood is the average likelihood over all mappings
**********************************************************************/
enum paramType {GLOBAL_RATE, RELATIVE_RATE, CHAR_MODEL_PARAM1, CHAR_MODEL_PARAM2, SEQ_MODEL_ASRV, HKY_MODEL_PARAM, GTR_MODEL_PARAM1, GTR_MODEL_PARAM2, GTR_MODEL_PARAM3, GTR_MODEL_PARAM4, GTR_MODEL_PARAM5, GTR_MODEL_PARAM6, TREE_LENGTH};
//enum combinedModelType {M0, M1}; //M0 = no rate difference between the binary characters. LL is computed based solely on the sequence data. M1= combined model

class charAndSeqModel {
public:
	charAndSeqModel(stochasticProcess* pCharModel, stochasticProcess* pSeqModel, MDOUBLE relRate, seqModelType nucModel, combinedModelType modelType, MDOUBLE treeLength);   
	charAndSeqModel(const charAndSeqModel &other);
	virtual ~charAndSeqModel();
	void copy(const charAndSeqModel *pOther);
	charAndSeqModel& operator=(const charAndSeqModel &other);
	virtual charAndSeqModel* clone() const {return new charAndSeqModel(*this);}
	void setRandomRateParams();
	
	stochasticProcess* getCharProcess() const {return _pCharModel;}
	stochasticProcess* getSeqProcess() const {return _pSeqModel;}
	void getSeqModelParams(vector<paramType>& params);
	void getCharModelParams(vector<paramType>& params);

	//MDOUBLE getGlobalRate() const {return _pSeqModel->getGlobalRate();}
	//void setGlobalRate(const MDOUBLE x) {_pSeqModel->setGlobalRate(x);}
	MDOUBLE getRelativeRate() const {return _relRate;}
	void setRelativeRate(const MDOUBLE x) {_relRate = x;}
	MDOUBLE getTreeLength() const {return _treeLength;}
	void setTreeLength(const MDOUBLE x) {_treeLength = x;}

	MDOUBLE getParam(paramType type) const;
	void setParam(paramType type, MDOUBLE val);
	string getParamName(paramType type) const;
	virtual MDOUBLE getParamLowerBound(paramType type) const; 
	virtual MDOUBLE getParamUpperBound(paramType type) const;
	void printModelParams(ostream &out) const;

	seqModelType getSeqModelType() const {return _nucModel;}
	void setParam2Bound (MDOUBLE val) {_param2Bound = val;}
	void setParam1Bound (const sequenceContainer& charContainer);
	void setTreeLengthBound (MDOUBLE val) {_treeLengthBound = val;};
	
	//*computes the combined model likelihood, which involves stochastic mapping
	MDOUBLE compLogLikelihood(const tree& inTree, const sequenceContainer& charContainer, const sequenceContainer& seqContainer, int mappingNum, int heuristicGroups, const vector<tree> &charTrees, llCalc llType);
	//*computes the combined model likelihood, which involves stochastic mapping, based on a vector of possible trees
	MDOUBLE compLogLikelihood(const vector<tree> &trees, const sequenceContainer& charContainer, const sequenceContainer& seqContainer, int mappingNum, int heuristicGroups, const vector<tree> &charTrees, llCalc llType);

	//*
	MDOUBLE compLogLikelihood(const tree& inTree, const sequenceContainer& seqContainer, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType);
	//*computes the likelihood of the sequence model based on a vector of stochastic mappings
	MDOUBLE compLogLikelihood(const tree& inTree, const vector<treeMutationHistory>& mappings, const sequenceContainer &seqContainer, int heuristicGroups, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType);
	//*computes the likelihood of the sequence model based on a vector of stochastic mappings - one for each tree in the input trees
	MDOUBLE compLogLikelihood(const vector<tree> &trees, const vector<treeMutationHistory>&allMappings, const sequenceContainer &seqContainer, int heuristicGroups, const vector<tree> &CharTrees, const sequenceContainer &charContainer, llCalc llType);

	//*computes the likelihood of the sequence model only based on a vector of trees that are the result of the stochastic mapping
	MDOUBLE compLogLikelihood(const vector<tree> &seqTrees, const sequenceContainer &seqContainer, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType);
	//*computes the likelihood of the sequence model based on a vector of trees that are the result of the stochastic mapping, and according to some heurisitc criterion. 
	MDOUBLE compLogLikelihood(const vector<tree> &trees, const sequenceContainer &seqContainer, int heuristicGroups, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType);

	//computes the likelihood of the sequence model only based on a vector of trees that are the result of the stochastic mapping AND weight for each tree
	MDOUBLE compLogLikelihood(const vector<tree> &trees, const sequenceContainer &seqContainer, const Vdouble& weights, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType);
	//*heuristic LL computation: get the average scaled tree from all mappings and computes LL based on the averaged scaled tree
	MDOUBLE compLogLikelihoodAvgTree(const vector<tree>& scaledTrees, const sequenceContainer& seqContainer, const vector<tree> &charTrees, const sequenceContainer &charContainer, llCalc llType);
	


	void scaleTrees(const tree& inTree, const vector<treeMutationHistory>& mappings, MDOUBLE relRate, vector<tree> &outtrees) const; 
	//scale multiple trees according to the corresponding mapping
	void scaleTrees(const vector<tree>& trees, const vector<treeMutationHistory>& mappings, MDOUBLE relRate, vector<tree> &outtrees) const; 
	tree getAvgScaledTree(const tree& inTree, const sequenceContainer& charContainer, int mappingNum);
	tree getAvgScaledTree(const vector<tree>& scaledTrees);

	static void getKMeanTrees(const vector<tree> &trees,  int k, vector<tree>& centerTrees, Vdouble& weights);

	combinedModelType getCombinedModelType() {return _modelType;};
	void setCombinedModelType(combinedModelType type) {_modelType = type;};

private:
	doubleRep getTreeLikelihoodAllPos(const tree& et, const sequenceContainer& sc, const stochasticProcess& sp, const Vdouble * const weights = NULL);


private:
	stochasticProcess* _pCharModel;
	stochasticProcess * _pSeqModel;
	MDOUBLE _relRate; 
	seqModelType _nucModel;
	combinedModelType _modelType;
	MDOUBLE _treeLength;

	MDOUBLE _param2Bound; //set multiplier for the upper and lower bound for characterModel param2
	MDOUBLE _param1Bound; //a lower bound for character model param 1. Accordingly, The upper bound is (1 - _param1Bound)
	MDOUBLE _treeLengthBound; //set multiplier for the upper and lower bound for total tree length

};
#endif
