#include "kMean.h"
#include "errorMsg.h"
#include "talRandom.h"
#include "matrixUtils.h"

using namespace std;


//returns the centers and the number of points belonging to each center
void kMean::run (const VVdouble& points, int k, int iters)
{
	if (k > points.size()) 
		errorMsg::reportError("cannot run kMean. Number of centers is higher than number of points");
	init(points.size(), k);
	calcDistances(points);
	if (k == 1)
	{
		chooseOnePoint(points);
		return;
	}
	chooseInitCenters(points , k);
	runKMeans(points, iters);
}


void kMean::init(int pointsNum, int k)
{
	resizeMatrix(_distances, pointsNum, pointsNum);
	_pointAssignments.resize(pointsNum);
	_centers.resize(k);
	_clusterAssignments.resize(k);
	_id2cluster.resize(pointsNum);
}

//calc the distance between any 2 points. 
//Each distance is the sum of squares over all dimensions
void kMean::calcDistances(const VVdouble& points)
{
	int pointsNum = points.size();
	int dim = points[0].size();
	for (int i = 0; i < pointsNum; ++i) {
		_distances[i][i] = 0.0;
		for (int j = i+1; j < pointsNum; ++j) {
			MDOUBLE sum = 0.0;
			for (int p = 0; p < dim; ++p)
				sum += (points[i][p] - points[j][p]) * (points[i][p] - points[j][p]);
			_distances[i][j] = sum;
			_distances[j][i] = sum;
		}
	}
}


//Chooses k centers from given points as follows:
//One center is chosen randomly.
//  - Now repeat k-1 times:
//    Add a new center x with probability proportional to the distance from x to the closest existing center
void kMean::chooseInitCenters(const VVdouble& points, int k)
{
	int pointsNum = points.size();
	Vdouble closestDist(pointsNum, 0.0); //the distance for each point to its closest center
	
	// Choose one random center and set the closestDist values
	int index = talRandom::giveIntRandomNumberBetweenZeroAndEntry(pointsNum);
    _centers[0] = index;
	MDOUBLE sumDist = 0.0;
    for (int i = 0; i < pointsNum; ++i) 
	{
		closestDist[i] = _distances[index][i];
		sumDist += closestDist[i];
    }
	
	//iterate k-1 times until all centers are set
	for (int c = 1; c < k; ++c)
	{
		//choose the new center based on the distances
		int centerIndex = getNewCenterBasedOnDistances(closestDist);
		_centers[c] = centerIndex;
		//calculate new closest distances
		for (int p = 0; p < pointsNum; ++p)
            closestDist[p] = min(_distances[p][centerIndex], closestDist[p]);
	}
}


int kMean::getNewCenterBasedOnDistances(const Vdouble& closestDist)
{
	MDOUBLE sumDist = 0.0;
	for (int i=0; i<closestDist.size(); i++)
		sumDist += closestDist[i];
	//choose a new center based on the calculated distances
	MDOUBLE ran = talRandom::giveRandomNumberBetweenZeroAndEntry(sumDist);
	MDOUBLE sum = 0.0;
    
	for (int loop = 0 ; loop < 10000; ++loop) 
	{
		for (int p=0; p < closestDist.size(); ++p) 
		{
			sum += closestDist[p];
			if (ran < sum)
				return p;
		}
	}
	LOGnOUT(5, <<"randon number = "<<ran<<" sumDist = "<<sumDist<<" sum = "<<sum<<endl);
	errorMsg::reportError("getNewCenterBasedOnDistances: could not give center. The reason is unknown.");
	return -1;
}

void kMean::runKMeans(const VVdouble& points, int numIterations)
{
	bool bChanged = assignPoints2Centers(points);
	for (int it = 0; it < numIterations; ++it)
	{
		if (bChanged == false)
			break;
		findNewCenters(points);
		bChanged = assignPoints2Centers(points);
	}
}


//Given pre-calculated cluster centers, stored in _centers, assign each point to its closest center
//Values are stored in _clusterAssignments
//returns true if one of the new assignments is different from the previous assignment
bool kMean::assignPoints2Centers(const VVdouble& points)
{
	bool bChanged = false;
	int k = _centers.size();
	_clusterAssignments.clear();	
	_clusterAssignments.resize(k);
	for (int p = 0; p < points.size(); ++p)
	{
		MDOUBLE minDist = VERYBIG;
		int bestClusterIndex = -1;
		for (int c = 0; c < k; ++c)
		{
			MDOUBLE dist = _distances[p][_centers[c]];
			if (dist < minDist)
			{
				minDist = dist;
				bestClusterIndex = c;
			}
		}
		_clusterAssignments[bestClusterIndex].push_back(p);
		if (_id2cluster[p] != _centers[bestClusterIndex]) {
			bChanged = true;
			_id2cluster[p] = _centers[bestClusterIndex];
		}
	}
	return bChanged;
}

//given a partition of the data points to k clusters (stored in _clusterAssignments) find a center for each cluster
//the new center is the point with the minimum distance to all other points in its cluster
void kMean::findNewCenters(const VVdouble& points)
{
	for (int c = 0; c < _centers.size(); ++c)
	{
		MDOUBLE minSumDist = VERYBIG;
		for (int p1 = 0; p1 < _clusterAssignments[c].size(); ++p1)
		{
			int index1 = _clusterAssignments[c][p1];
			MDOUBLE sum = 0.0;
			for (int p2 = 0; p2 < _clusterAssignments[c].size(); ++p2)
			{
				int index2 = _clusterAssignments[c][p2];
				MDOUBLE dist = _distances[index1][index2];
				sum += dist;
			}
			if (sum < minSumDist) 
			{
				_centers[c] =  index1;
				minSumDist = sum;
			}
		}

	}
}

void kMean::chooseOnePoint(const VVdouble& points)
{
	MDOUBLE minSumDist = VERYBIG;
	for (int p1 = 0; p1 < points.size(); ++p1)
	{
		MDOUBLE sum = 0.0;
		for (int p2 = 0; p2 < points.size(); ++p2)
			sum += _distances[p1][p2];
		if (sum < minSumDist) 
		{
			_centers[0] =  p1;
			minSumDist = sum;
		}
		_clusterAssignments[0].push_back(p1);
	}
	
}
