#include "simulateCombinedModel.h"
#include "talRandom.h"
#include "matrixUtils.h"
#include "treeIt.h"
#include "evaluateCharacterFreq.h"
#include <algorithm>


simulateCombinedModel::simulateCombinedModel(const tree& inTree, const charAndSeqModel* pCombinedModel, const alphabet* pCharAlph, const alphabet* pSeqAlph)
: _tree(inTree), _pCombinedModel(pCombinedModel), _pCharAlph(pCharAlph), _pSeqAlph(pSeqAlph), _pSeqSimulator(NULL)
{
}

simulateCombinedModel::~simulateCombinedModel()
{
	if (_pSeqSimulator != NULL)
		delete _pSeqSimulator;
}

void simulateCombinedModel::init()
{
	//init the vector of waiting times. 
	twoStateModel* pCharModel = static_cast<twoStateModel*>(_pCombinedModel->getCharProcess()->getPijAccelerator()->getReplacementModel());
	_waitingTimeParams.clear();
	_waitingTimeParams.resize(pCharModel->alphabetSize());
	int i, j;
	for (i = 0; i < pCharModel->alphabetSize(); ++i)
	{
		_waitingTimeParams[i] = -pCharModel->dPij_dt(i, i, 0.0);
	}

	//init _jumpProbs. _jumpProbs[i][j] = Q[i][j] / -Q[i][i]
	_jumpProbs.clear();
	_jumpProbs.resize(pCharModel->alphabetSize());
	for (i = 0; i < pCharModel->alphabetSize(); ++i)
	{
		MDOUBLE sum = 0.0;
		_jumpProbs[i].resize(pCharModel->alphabetSize());
		for (j = 0; j < pCharModel->alphabetSize(); ++j)
		{
			if (i == j)
				_jumpProbs[i][j] = 0.0;
			else
			{
				_jumpProbs[i][j] = pCharModel->dPij_dt(i, j, 0.0) / _waitingTimeParams[i];
			}
			sum += _jumpProbs[i][j];
		}
		if (! DEQUAL(sum, 1.0)){
			string err = "error in simulateCombinedModel::init(): sum probabilities is not 1 and equal to " + double2string(sum);
			errorMsg::reportError(err);
		}
	}
}


void simulateCombinedModel::simulate(int seqPosNum, const sequenceContainer* pOrigCharContainer, MDOUBLE overlap)
{
	assert(_pCharAlph->size() == 2); //needed just for the overlap calculations since assuming that overlap of 0.1 is actually 0.9 if flipping the characters
	init();
	treeMutationHistory mappings;
	simulateCharacterModel(mappings, pOrigCharContainer, overlap);
	vector<treeMutationHistory> mappingVec;
	mappingVec.push_back(mappings);
	vector<tree> scaledTree;
	_pCombinedModel->scaleTrees(_tree, mappingVec, _pCombinedModel->getRelativeRate(), scaledTree);

    if (_pSeqSimulator)
		delete _pSeqSimulator;
	_pSeqSimulator = new simulateTree(scaledTree[0], *_pCombinedModel->getSeqProcess(), _pSeqAlph);
	_pSeqSimulator->generate_seq(seqPosNum);

	string outTree = traitRateOptions::_outDir + "//simTree.tree";
	scaledTree[0].output(outTree);
}

void simulateCombinedModel::simulateCharacterModel(treeMutationHistory& res, const sequenceContainer* pOrigCharContainer, MDOUBLE overlapThreshold)
{
	_nodesContent.resize(_tree.getNodesNum());
	twoStateModel* pCharModel = static_cast<twoStateModel*>(_pCombinedModel->getCharProcess()->getPijAccelerator()->getReplacementModel());
	Vdouble freqs(pCharModel->alphabetSize(),0.0);
	for (int i = 0; i< freqs.size(); ++i)
		freqs[i]= pCharModel->freq(i);

	int maxTries = 1000;
	int it = 0;
	for (; it < maxTries; ++it)
	{
		treeMutationHistory trySim(_tree.getNodesNum());
		int rootState = giveRandomState(_pCharAlph->size(), freqs);
		string name = _tree.getRoot()->name();
		sequence seq(_pCharAlph->fromInt(rootState),name, "", _tree.getRoot()->id(), _pCharAlph);
		_charContainer.add(seq);

		_nodesContent[_tree.getRoot()->id()] = rootState;
		for (int s = 0; s < _tree.getRoot()->getNumberOfSons(); ++s)
			simulateSon(_tree.getRoot(),0,rootState, s, trySim);
		Vint type4Pos;
		sequenceContainer leavesSc = toCharDataWithoutInternalNodes();
		getCharacterType4pos(leavesSc, type4Pos);
		if (type4Pos[0] > 1)
		{
			if (overlapThreshold > 0.5)
			{
                //calc overlap between simulated and real characters
				int eq = getIdenticalSeqs(leavesSc, *pOrigCharContainer, 0);
				MDOUBLE overlap = static_cast<MDOUBLE>(eq) / leavesSc.numberOfSeqs();
				if (overlap < 0.5)
					overlap = 1 - overlap; 
				if (overlap < overlapThreshold) {
					res = trySim;
					break;
					}
			}
			else {
				res = trySim;
				break;
			}
		}
		_charContainer.removeAll(); //in case the simulation did not succeed
	}
	if (it == maxTries)
		errorMsg::reportError("could not produce only one character type found in the character data");
}


//simulate jumps starting from curNode (with state = previousContent) along the branch leading to whichSon. 
//disFromNode is the distance from the current node (in some cases we don't start the simulation form the node but some distance "down" from it)
//In each step:
//1. Draw a new waiting time = timeTillChange.
//2. if the total distance from the node (timeTillChange+disFromNode) is smaller than the branch length then 
//		we draw a new waiting time and update disFromNode and repeat
//   otherwise (the total distance passed the son node)
//		For all but the first son we start the simlation from the begining of its edge.
//		For the first son: the jump passed the begining of the edge, so we should first check how many descendents the jump passed, and then start the simulation thatdistance away from the begining of the edge
void simulateCombinedModel::simulateSon(tree::nodeP curNode, MDOUBLE disFromNode, int previousContent, int whichSon, treeMutationHistory& res)
{
	tree::nodeP sonNode = curNode->getSon(whichSon);
	MDOUBLE avgWaitingTime = 1.0 / _waitingTimeParams[previousContent];
	MDOUBLE timeTillChange = talRandom::rand_exp(avgWaitingTime);
	disFromNode += timeTillChange;
	int nextContent = giveRandomState(_pCharAlph->size(), previousContent, _jumpProbs);
	//loop until we reach the son node
	while (disFromNode < sonNode->dis2father()) 
	{
		pair<int, MDOUBLE> oneJump(previousContent, timeTillChange);
		res[sonNode->id()].push_back(oneJump);
		previousContent=nextContent;
		MDOUBLE avgWaitingTime = 1.0 / _waitingTimeParams[previousContent];
		timeTillChange = talRandom::rand_exp(avgWaitingTime);
		disFromNode += timeTillChange;
		nextContent = giveRandomState(_pCharAlph->size(), nextContent, _jumpProbs);
	}
	//we now simulate all sons of the son node.
	//but - the last jump may "passed" several internal nodes. 
	while (disFromNode >= sonNode->dis2father()) 
	{
		_nodesContent[sonNode->id()] = previousContent;
		string name = sonNode->name();
		sequence seq(_pCharAlph->fromInt(previousContent),name, "", sonNode->id(), _pCharAlph);
		_charContainer.add(seq);

		MDOUBLE timeOfJumpTillSon = timeTillChange - (disFromNode - sonNode->dis2father());
		pair<int, MDOUBLE> oneJump(previousContent, timeOfJumpTillSon);
		res[sonNode->id()].push_back(oneJump);

		if (sonNode->isLeaf()) {
			return;
		}
		for (int s = 1; s < sonNode->getNumberOfSons(); ++s)
            simulateSon(sonNode, 0, previousContent, s, res);
		disFromNode -= sonNode->dis2father();
		timeTillChange -= timeOfJumpTillSon;
		curNode = sonNode;
		sonNode = curNode->getSon(0);
	}
	pair<int, MDOUBLE> oneJump(previousContent, timeTillChange);
	res[sonNode->id()].push_back(oneJump);

	simulateSon(curNode, disFromNode, nextContent, 0, res);
}


sequenceContainer simulateCombinedModel::toCharDataWithoutInternalNodes() const {
	sequenceContainer myseqData;
	int nextId = 0;
	sequenceContainer::constTaxaIterator tIt;
	sequenceContainer::constTaxaIterator tItEnd;
	tIt.begin(_charContainer);
	tItEnd.end(_charContainer);

	for (; tIt!= tItEnd; ++tIt)
	{
		string nodeName = tIt->name();
		tree::nodeP theCurNode = _tree.findNodeByName(nodeName);
		if (theCurNode == NULL)
			errorMsg::reportError("could not find the specified name: " + nodeName);
		if (theCurNode->isInternal()) 
			continue;
		sequence tmpSeq = *tIt;
		tmpSeq.setID(nextId);
		myseqData.add(tmpSeq);
		++nextId;
	}
	return myseqData;
}

sequenceContainer simulateCombinedModel::toSeqDataWithoutInternalNodes() const
{
	if (!_pSeqSimulator)
		errorMsg::reportError("error in simulateCombinedModel::toSeqDataWithoutInternalNodes _pSeqSimulator is NULL");
	return _pSeqSimulator->toSeqDataWithoutInternalNodes();
}

int simulateCombinedModel::getIdenticalSeqs(const sequenceContainer& sc1, const sequenceContainer& sc2, int pos)
{
	int res = 0;
	for(int i = 0; i < sc1.numberOfSeqs(); ++i)
	{
		bool bFound = false;
        for (int j = 0; j < sc2.numberOfSeqs(); ++j)
		{
		    if(sc1[i].name() == sc2[j].name())
			{
				if (sc1[i][pos] == sc2[j][pos]) {
					++res;
                    break;
				}
			}
		}
		if (bFound == false) 
		{
			string msg = string("Can't find sequence name in the second MSA: ") + sc1[i].name();
            errorMsg::reportError(msg);
			return false;
		}
	}
	return true;
}

