#include "stochasticMapping.h"
//#include "treeUtil.h"
#include "talRandom.h"
#include "computePijComponent.h"
#include "computeUpAlg.h"
#include "treeIt.h"
#include "matrixUtils.h"

stochasticMapping::stochasticMapping(const tree& inTree, const stochasticProcess& sp)
: _et(inTree), _sp(sp)
{
	init();
}

stochasticMapping::~stochasticMapping() 
{}

void stochasticMapping::init() 
{
	_waitingTimeParams.clear();
	_waitingTimeParams.resize(_sp.alphabetSize());
	int i, j;
	for (i = 0; i < _sp.alphabetSize(); ++i)
	{
		_waitingTimeParams[i] = -_sp.dPij_dt(i, i, 0.0);
		
	}

	//init _jumpProbs.
	//_jumpProbs[i][j] = Q[i][j] / -Q[i][i]
	_jumpProbs.clear();
	_jumpProbs.resize(_sp.alphabetSize());
	for (i = 0; i < _sp.alphabetSize(); ++i)
	{
		MDOUBLE sum = 0.0;
		_jumpProbs[i].resize(_sp.alphabetSize());
		for (j = 0; j < _sp.alphabetSize(); ++j)
		{
			if (i == j)
				_jumpProbs[i][j] = 0.0;
			else
			{
				_jumpProbs[i][j] = _sp.dPij_dt(i, j, 0.0) / _waitingTimeParams[i];
			}
			sum += _jumpProbs[i][j];
		}
		if (! DEQUAL(sum, 1.0)){
			string err = "error in stochasticMapping::init(): sum probabilities is not 1 and equal to ";
			err+=double2string(sum);
			errorMsg::reportError(err);
		}
	}
	_nodesContent.clear();
	_nodesContent.resize(_et.getNodesNum(), 0);
}

//simulate mutations on the phylogeny so the resulting leaves are identical to the input data, 
//as given in the specified position of the sequenceContainer. 
//The result is stored as the history of mutations for each branch.
//retruns false if fails to generate mapping for the specified data
bool stochasticMapping::generateStochasticMapping(const sequenceContainer& inSc, int pos, int iterations, vector<treeMutationHistory >& res)
{
	//1. compute conditional probabilities
	computePijHom pij;
	pij.fillPij(_et, _sp, 0, _sp.isReversible()); 	

	computeUpAlg cup;
	suffStatGlobalHomPos sscUp;
	sscUp.allocatePlace(_et.getNodesNum(), _sp.alphabetSize());
	cup.fillComputeUp(_et, inSc, pos, pij, sscUp);

	//2. compute root probs and sampling probs given data and ancestral state
	Vdouble rootProbs = getRootProbs(sscUp);
	VVVdouble samplingProbs = getSamplingProbs(sscUp);

	res.resize(iterations);
	for (int i = 0; i < iterations; ++i)
	{
		//3. sample ancestrals and store in _nodeContants
        sampleAncestrals(rootProbs, samplingProbs);
        //4. simulate mutations based on ancestrals
        sampleMutationsGivenAncestrals(res[i]);
	}
	return true;
}

Vdouble stochasticMapping::getRootProbs(const suffStatGlobalHomPos& sscUp) const
{
	Vdouble rootProbs(_sp.alphabetSize(),0.0);
	MDOUBLE sum = 0.0;
	int let = 0;
	for (; let< rootProbs.size(); ++let)
	{
		rootProbs[let] = _sp.freq(let) * convert(sscUp.get(_et.getRoot()->id(),let));
		sum += rootProbs[let];
	}
	for (let = 0; let< rootProbs.size(); ++let)
		rootProbs[let] /= sum;
	return rootProbs;
}


VVVdouble stochasticMapping::getSamplingProbs(const suffStatGlobalHomPos& sscUp) const
{
	//compute sampling probabilities for each node: P(letter at Node is j | Data & letter at Father(Node) is i)
	//P(Node=j | Data, Father=i) = Up[node][j] * Pij(t) / Sum_k {Up[node][k] * Pik(t) }
	VVVdouble samplingProbs(_et.getNodesNum());//samplingProbs[nodeId][fatherState][sonState]
	treeIterTopDownConst tIt(_et);
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
	{
		if (mynode->isRoot())
			continue;
		resizeMatrix(samplingProbs[mynode->id()], _sp.alphabetSize(), _sp.alphabetSize());
		for (int fatherState = 0; fatherState < _sp.alphabetSize(); ++fatherState)
		{
			MDOUBLE sum = 0.0; 
			for (int sonState = 0; sonState < _sp.alphabetSize(); ++sonState)
			{
				doubleRep tmp = _sp.Pij_t(fatherState, sonState, mynode->dis2father()) * sscUp.get(mynode->id(),sonState);
				samplingProbs[mynode->id()][fatherState][sonState] = convert(tmp);
				sum += convert(tmp);
			}
			for (int sonState = 0; sonState < _sp.alphabetSize(); ++sonState)
				samplingProbs[mynode->id()][fatherState][sonState] /= sum;
		}
	}
	return samplingProbs;


}

void stochasticMapping::sampleAncestrals(const Vdouble& rootProbs, const VVVdouble& samplingProbs)
{
	treeIterTopDownConst tIt(_et);
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
	{
		int nodeId = mynode->id();
		if (mynode->isRoot())
		{
			int rootState = giveRandomState(_sp.alphabetSize(), rootProbs);
			_nodesContent[nodeId] = rootState;
		}
		else
		{
			int fatherState = _nodesContent[mynode->father()->id()];
			int sonState = giveRandomState(_sp.alphabetSize(), fatherState, samplingProbs[nodeId]);
			_nodesContent[nodeId] = sonState;
		}
	}
}

void stochasticMapping::sampleMutationsGivenAncestrals(treeMutationHistory& res)
{
	res.resize(_et.getNodesNum());
	treeIterTopDownConst tIt(_et);
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
	{
		int nodeId = mynode->id();
		if (mynode->isRoot())
			continue;
		else
		{
			int fatherState = _nodesContent[mynode->father()->id()];
			int sonState = _nodesContent[nodeId];
			res[nodeId] = sampleMutationsGivenAncestralsPerBranch(fatherState, sonState, mynode->dis2father());
		}
	}
}


mutationHistoryPerBranch stochasticMapping::sampleMutationsGivenAncestralsPerBranch(int fatherState, int sonState, MDOUBLE branchLength)
{
	int maxIterNum = 1000;
	for (int i = 0; i < maxIterNum; ++i)
	{
		mutationHistoryPerBranch tryMapping;
		MDOUBLE disFromNode = 0.0;
		int curState = fatherState;
		MDOUBLE avgWaitingTime = 1.0 / _waitingTimeParams[curState];
		MDOUBLE timeTillChange = talRandom::rand_exp(avgWaitingTime);

		if (fatherState != sonState)
		{ //draw jump time given a change has occured
			MDOUBLE u = talRandom::giveRandomNumberBetweenZeroAndEntry(1.0);
            MDOUBLE tmp = u * (1.0 - exp(branchLength * -_waitingTimeParams[curState]));
            timeTillChange =  -log(1.0 - tmp) / _waitingTimeParams[curState];
			assert (timeTillChange < branchLength);
		}
        
		while (disFromNode + timeTillChange < branchLength)
		{
            //a jump occured but not passed the whole branch. Add the current state and time to branch history and draw next state
			pair<int, MDOUBLE> oneJump(curState, timeTillChange);
			tryMapping.push_back(oneJump);
			curState = giveRandomState(_sp.alphabetSize(), curState, _jumpProbs);
			disFromNode += timeTillChange;			
			avgWaitingTime = 1.0 / _waitingTimeParams[curState];
            timeTillChange = talRandom::rand_exp(avgWaitingTime);
       	}
		//the last jump passed the length of the branch
		if (curState != sonState)
		{ //simulation failed
			continue;
		}
		else
		{
			MDOUBLE timeOfJump = branchLength - disFromNode;
			pair<int, MDOUBLE> oneJump(curState, timeOfJump);
			tryMapping.push_back(oneJump);
			return tryMapping;
		}
	}
	//all simulations failed
	string err = "could not produce simulations with father = " + int2string(fatherState) + " son = " + int2string(sonState) + "branch length = " + double2string(branchLength);
	errorMsg::reportError(err);
	mutationHistoryPerBranch stam;
	return stam;

}

vector<treeMutationHistory> stochasticMapping::getMappingForEachTree(const vector<tree>& trees, const stochasticProcess* pCharModel, const sequenceContainer& inSc, int pos)
{
	vector<treeMutationHistory> res;
	for (int t = 0; t < trees.size(); ++t)
	{
		stochasticMapping sm(trees[t], *pCharModel);
		vector<treeMutationHistory> oneMappng;
		sm.generateStochasticMapping(inSc, pos, 1, oneMappng);
		res.push_back(oneMappng[0]);
	}
	return res;
}
