#include "definitions.h"
#include "traitRateOptions.h"
#include "traitRateMng.h"
#include "talRandom.h"
#include "errorMsg.h"
#include "utils.h"
#include "treeUtil.h"
#include "recognizeFormat.h"
#include "fastaFormat.h"
#include "nucleotide.h"
#include "AddLog.h"
#include <fstream>
#include <string>
#include <algorithm>
using namespace std; 

void rootTreeAccording2ModelTree(int argc, char **argv);
void getNexusTrees(int argc, char **argv);
void compareExternalBranches(int argc, char **argv);
void extractCodonPosFromMSA(int argc, char **argv);
void concatenateMSA(int argc, char **argv);
void calcLLFromDistribution();

int main(int argc, char **argv) {
	
	talRandom::setSeed(1);
	//getNexusTrees(argc, argv);
	//return 1;
	

	if (argc < 2)
		errorMsg::reportError("No input arguements. Please specify path to parameter file");
	string paramStr = argv[1];
	traitRateOptions::initOptions(paramStr);

	if (traitRateOptions::_mainType == "simulateSMIterations")
	{
		traitRateMng mng;
		mng.simulateSMIterations();
		return 0;
	}
	if (traitRateOptions::_mainType == "simulateRRFixedTaxa")
	{
		traitRateMng mng;
		mng.simulateRRFixedTaxa();
		return 0;
	}
	if (traitRateOptions::_mainType == "simulateNullModel")
	{
		traitRateMng mng;
		mng.simulateNullModel();
		return 0;
	}
	if (traitRateOptions::_mainType == "simulateFalsePositive")
	{
		traitRateMng mng;
		mng.simulateFalsePositive();
		return 0;
	}
	if (traitRateOptions::_mainType == "calcLLDistribution")
	{
		traitRateMng mng;
		mng.calcLLDistribution();
		return 0;
	}
	if (traitRateOptions::_mainType == "calcLLFromDistribution")
	{
		calcLLFromDistribution();
		return 0;
	}

	
	traitRateMng mng;
    mng.run();
	return 0;
}

void rootTreeAccording2ModelTree(int argc, char **argv)
{
	if (argc != 4)
		errorMsg::reportError("rootTreeAccording2ModelTree USAGE: <inModelTree> <inUnrootedTree> <outRooted> ");
	string inModelTree = argv[1];
	string inUnrootedTree = argv[2];
	string outTree = argv[3];

	makeBifurcatingTree(inModelTree, inUnrootedTree, outTree);
}

void getNexusTrees(int argc, char **argv)
{
	if (argc != 5)
		errorMsg::reportError("getTrees USAGE: <inTrees file 1> <inTrees file 2> <out trees> <% burn-in");
	string inTreesFile1 = argv[1];
	string inTreesFile2= argv[2];
	string outTrees = argv[3];
	MDOUBLE burnIn = string2double(argv[4]);

	vector<tree> trees1 = getNexusTreesFromFile (inTreesFile1);
	vector<tree> trees2 = getNexusTreesFromFile (inTreesFile2);
	ofstream out(outTrees.c_str());
	int cutoff = static_cast<int>(trees1.size() * burnIn);
	for (int t = cutoff; t < trees1.size(); ++t) {
		trees1[t].output(out);
	}
	cutoff = static_cast<int>(trees2.size() * burnIn);
	for (int t = cutoff; t < trees2.size(); ++t) {
		trees2[t].output(out);
	}

	out.close();
}

void compareExternalBranches(int argc, char **argv)
{
	if (argc != 3)
		errorMsg::reportError("compareExternalBranches USAGE: <inTrees file 1> <inTrees file 2> <out trees> <% burn-in");
	string inTreesFile1 = argv[1];
	string inTreesFile2= argv[2];

	tree tree1(inTreesFile1.c_str());
	tree tree2(inTreesFile2.c_str());

	MDOUBLE sum1 = getSumTipLengths(tree1);
	MDOUBLE sum2 = getSumTipLengths(tree2);

	cerr<<"sum1="<<sum1<<" sum2="<<sum2<<endl;
}


void extractCodonPosFromMSA(int argc, char **argv)
{
	if (argc != 4)
		errorMsg::reportError("extractCodonPosFromMSA USAGE: <in MSA> <OUT MSA> <codon position>");
	string inMsaFileName = argv[1];
	string outMsaFileName= argv[2];
	int pos = atoi(argv[3]);

	nucleotide alph;
	ifstream in(inMsaFileName.c_str());
	sequenceContainer inSc = recognizeFormat::read(in, &alph);
	in.close();

	sequenceContainer outSc = extractCodonPos(inSc, pos);
	ofstream outMsaF(outMsaFileName.c_str());
	fastaFormat::write(outMsaF, outSc);
	outMsaF.close();
}

void concatenateMSA(int argc, char **argv)
{
	if (argc != 4)
		errorMsg::reportError("concatenateMSA USAGE: <in MSA1> <in MSA2> <OUT MSA>");
	string inMsa1FileName = argv[1];
	string inMsa2FileName = argv[2];
	string outMsaFileName= argv[3];
	
	nucleotide alph;
	ifstream in1(inMsa1FileName.c_str());
	sequenceContainer inSc1 = recognizeFormat::read(in1, &alph);
	in1.close();
	ifstream in2(inMsa2FileName.c_str());
	sequenceContainer inSc2 = recognizeFormat::read(in2, &alph);
	in2.close();

	bool bok = inSc1.concatenate(inSc2);
	if (!bok)
		errorMsg::reportError("cannot concatenate the 2 MSAs");


	ofstream outMsaF(outMsaFileName.c_str());
	fastaFormat::write(outMsaF, inSc1);
	outMsaF.close();

}

void calcLLFromDistribution()
{
	string outDirName = traitRateOptions::_outDir;
	int computeLLNum = traitRateOptions::_stochasicMappingIterations;
	/////////////////////////////////////
	//read LL file and store in llVec1iter
	string llFile = traitRateOptions::_seqFile;
	ifstream in(llFile.c_str());
	vector<string> lines;
	putFileIntoVectorStringArray(in, lines);
	in.close();
	Vdouble llVec1iter;
	vector<string>::const_iterator it1 = lines.begin();
	while ( ( (*it1).find("ITER")  == -1) && (it1 != lines.end()))
	{ 
		++it1;
	}
	//++it1; //skip header 
	for (; it1 != lines.end(); ++it1)
	{
		if (it1->empty() || (takeCharOutOfString("\t ", *it1).empty()))
			continue;
		//if (((*it1).find("ITER")) || ((*it1).find("AVG")) || ((*it1).find("STD")))
		if ( (it1->find("ITER")  != -1) || (it1->find("AVG")  != -1) || (it1->find("STD")  != -1))
			continue;
        vector<string> strVec;
		splitString(*it1, strVec, "\t "); 
		MDOUBLE ll = string2double(strVec[1]);
		doubleRep llrep(ll);
		//doubleRep l = exp(llrep);
		llVec1iter.push_back(ll);
		//lVec.push_back(l);
	}
	/////////////////////////////////////
	//calc average LL for the specified number of iterations
	int llNum = llVec1iter.size();

    for (int iterNum = traitRateOptions::_startSimulationsIter; iterNum <= traitRateOptions::_endSimulationsIter; iterNum*= 2)
	{
		string llFileName = outDirName + "//" + int2string(iterNum) + "_LL.txt";
		ofstream LLfile4iter(llFileName.c_str());
		LLfile4iter<<"ITER"<<"\t"<<"LL"<<endl;

		Vdouble llVec(computeLLNum);
		for (int j = 0; j < computeLLNum; ++j)
		{
			MDOUBLE sumlog = -1;
			for (int i = 0; i < iterNum; ++i) 
			{
				int r = talRandom::giveIntRandomNumberBetweenZeroAndEntry(llNum);
				MDOUBLE li = llVec1iter[r];
				if (sumlog == -1)
					sumlog = li;
				else
                    sumlog = AddLog(li, sumlog);
			}
			//compute average l:
			MDOUBLE ll = sumlog - log(static_cast<MDOUBLE>(iterNum));
			llVec[j] = ll;
			LLfile4iter<<j<<"\t"<<ll<<endl;
		}
		MDOUBLE avg = computeAverage(llVec);
		MDOUBLE std = computeStd(llVec);
		sort(llVec.begin(), llVec.end());
		MDOUBLE dif = fabs(llVec[0] - llVec[llVec.size()-1]);
		LLfile4iter<<"AVG"<<"\t"<<avg<<endl;
		LLfile4iter<<"STD"<<"\t"<<std<<endl;
		LLfile4iter<<"DIF"<<"\t"<<dif<<endl;
		LLfile4iter.close();
	}
}
