#include "traitRateMng.h"
#include "traitRateOptions.h"
#include "traitRateOptimizer.h"
#include "simulateCombinedModel.h"
#include "utils.h"
#include "treeUtil.h"
#include "stochasticMapping.h"
#include "charAndSeqModel.h"
#include "recognizeFormat.h"
#include "logFile.h"
#include "Parameters.h"
#include "evaluateCharacterFreq.h"
#include "trivialAccelerator.h"
#include "uniDistribution.h"
#include "gammaDistribution.h"
#include "nucJC.h"
#include "gtrModel.h"
#include "hky.h"
#include "treeIt.h"
#include "seqContainerTreeMap.h"
#include "matrixUtils.h"
#include "fastaFormat.h"
#include "phylipFormat.h"
#include "nucleotide.h"
#include "amino.h"
#include "aaJC.h"
#include "readDatMatrix.h"
#include "chebyshevAccelerator.h"
#include <algorithm>
#include <cmath>
#include <ctime>
using namespace std;

traitRateMng::traitRateMng() {
	init();
}

traitRateMng::~traitRateMng() {
	clear();
}

void traitRateMng::clear() {
	if (_pCombinedModel != NULL)
	{
		delete _pCombinedModel;
		_pCombinedModel = NULL;
	}
	if (_pSeqAlph)
	{
		delete _pSeqAlph;
        _pSeqAlph = NULL;
	}
}


void traitRateMng::init() {
	_pCombinedModel = NULL;
	_pSeqAlph = NULL;
	_bestLL = 0;
	_model0LL = 0;
	createDir("", traitRateOptions::_outDir);
	string logFileName("");
	if (traitRateOptions::_logFile.size() > 0)
         logFileName = traitRateOptions::_outDir + "//" + traitRateOptions::_logFile;
	myLog::setLog(logFileName, traitRateOptions::_logValue);
}

void traitRateMng::run()
{
    printOptionParameters();
	getStartingData();
	getTree();
	validateInputs();
	initStartingModel();
	optimizeTreeAndInitialSeqModel();

	_bestLL = optimizeCombinedModel();
	printResults();
}


//simulate over iterNum trees different number of stochastic mapping iterations
void traitRateMng::simulateSMIterations()
{
	int startSMIters = 25;
	int endSMIters = 800;
	int numSMIters = 6;
	int computeLLNum = 100;
	
    printOptionParameters();
	//the model is the same for all iterations and doesn't change due to optimizations
	initAlphabet();
	initStartingModel();

	int iterNum = 1+ traitRateOptions::_endSimulationsIter - traitRateOptions::_startSimulationsIter;
	
	string baseOutDirName = traitRateOptions::_outDir;
	string avgLLfileName = baseOutDirName + "//avg.txt";
	string stdLLfileName = baseOutDirName + "//std.txt";
	string difLLfileName = baseOutDirName + "//dif.txt";
	ofstream avgLLfile(avgLLfileName.c_str());
	ofstream stdLLfile(stdLLfileName.c_str());
	ofstream difLLfile(difLLfileName.c_str());
	avgLLfile<<"TREE_ITER"<<"\t";
	stdLLfile<<"TREE_ITER"<<"\t";
	difLLfile<<"TREE_ITER"<<"\t";
		
	int sm, it;
	VVdouble stdMat, avgMat, difMat;
	resizeMatrix(stdMat, numSMIters, iterNum);  
	resizeMatrix(avgMat, numSMIters, iterNum);  
	resizeMatrix(difMat, numSMIters, iterNum);  
	for (sm = startSMIters; sm <= endSMIters; sm*=2)
	{
		avgLLfile<<sm<<"\t";
		stdLLfile<<sm<<"\t";
		difLLfile<<sm<<"\t";
	}
	avgLLfile<<endl;
	stdLLfile<<endl;
	difLLfile<<endl;
	
	for (it = traitRateOptions::_startSimulationsIter; it <= traitRateOptions::_endSimulationsIter; ++it)
	{
		LOGnOUT(3, <<"simulation iter = "<<it<<endl);
		//open simulated directory and out files
		string iterOutDirName = baseOutDirName + "//tree" + int2string(it);
		traitRateOptions::_outDir = iterOutDirName;
		createDir("", iterOutDirName);
		avgLLfile<<"tree_"<<it<<"\t";
		stdLLfile<<"tree_"<<it<<"\t";
		difLLfile<<"tree_"<<it<<"\t";
		
		//get tree and simulate 
		getSimulatedTree(it);
		simulateCombinedModel sim(_tree, _pCombinedModel, &_charAlph, _pSeqAlph);
		sim.simulate(traitRateOptions::_simulationSeqPos);
		sequenceContainer simCharContainer = sim.toCharDataWithoutInternalNodes();
		sequenceContainer simSeqContainer = sim.toSeqDataWithoutInternalNodes();
		string outCharName = iterOutDirName + "//simChars.fasta";
		string outSeqName = iterOutDirName + "//simSeqs.fasta";
		string outTreeName = iterOutDirName + "//model.tree";
		ofstream outChar(outCharName.c_str());
		ofstream outSeq(outSeqName.c_str());
		fastaFormat::write(outChar, simCharContainer);
		fastaFormat::write(outSeq, simSeqContainer);
		_tree.output(outTreeName);
        outChar.close();
		outSeq.close();

		//compute LL 100 times for different number of sm iterations
		sm = startSMIters;
		for (int itNum = 0; itNum < numSMIters; ++itNum, sm*=2)
		{
			string llFileName = iterOutDirName + "//" + int2string(sm) + "_LL.txt";
			ofstream LLfile4iter(llFileName.c_str());
			LLfile4iter<<"ITER"<<"\t"<<"LL"<<endl;

			Vdouble llVec(computeLLNum);
			for (int i = 0; i < computeLLNum; ++i)
			{
				_treeVec = traitRateOptimizer::pickTrees(_treeVec, sm);
				MDOUBLE l = _pCombinedModel->compLogLikelihood(_treeVec, simCharContainer, simSeqContainer, sm, 0,_treeVec, traitRateOptions::_llCalcType);
				llVec[i] = l;
				LLfile4iter<<i<<"\t"<<l<<endl;
			}
			LLfile4iter.close();
			MDOUBLE avg = computeAverage(llVec);
			MDOUBLE std = computeStd(llVec);
			sort(llVec.begin(), llVec.end());
			MDOUBLE dif = fabs(llVec[0] - llVec[llVec.size()-1]);
			avgLLfile<<avg<<"\t";
			stdLLfile<<std<<"\t";
			difLLfile<<dif<<"\t";
			stdMat[itNum][it] = std;
			avgMat[itNum][it] = avg;
			difMat[itNum][it] = dif;

		}
		avgLLfile<<endl;
		stdLLfile<<endl;
		difLLfile<<endl;
	}
	
	avgLLfile<<"AVG"<<"\t";
	stdLLfile<<"AVG"<<"\t";
	difLLfile<<"AVG"<<"\t";
	for (int it = 0; it < numSMIters; ++it)
	{
		avgLLfile<<computeAverage(avgMat[it])<<"\t";
		stdLLfile<<computeAverage(stdMat[it])<<"\t";
		difLLfile<<computeAverage(difMat[it])<<"\t";
	}
	avgLLfile<<endl;
	stdLLfile<<endl;
	difLLfile<<endl;
 	avgLLfile.close();
	stdLLfile.close();
	difLLfile.close();
}



void traitRateMng::calcLLDistribution()
{
    printOptionParameters();
	getStartingData();
	getTree();
	validateInputs();
	initStartingModel();

	string baseOutDirName = traitRateOptions::_outDir;
	string llFileName = traitRateOptions::_outDir + "//ll.txt";
	ofstream llfile(llFileName.c_str());
	llfile<<"ITER"<<"\t"<<"LL"<<"\t"<<"L"<<"\n";
	Vdouble llVec(traitRateOptions::_endSimulationsIter);
	Vdouble lVec(traitRateOptions::_endSimulationsIter);
	_treeVec = traitRateOptimizer::pickTrees(_treeVec, traitRateOptions::_stochasicMappingIterations);
	for (int it = 0; it < traitRateOptions::_endSimulationsIter; ++it) {
		MDOUBLE ll = _pCombinedModel->compLogLikelihood(_treeVec, _charContainer, _seqContainer, traitRateOptions::_stochasicMappingIterations, 0, _treeVec, traitRateOptions::_llCalcType);
		MDOUBLE l = exp(ll);
		llVec[it] = ll;
		lVec[it] = l;
		llfile<<it<<"\t"<<ll<<"\t"<<l<<endl;
	}
    MDOUBLE avg_ll = computeAverage(llVec);
	MDOUBLE std_ll = computeStd(llVec);
    MDOUBLE avg_l = computeAverage(lVec);
	MDOUBLE std_l = computeStd(lVec);
	llfile<<"AVG"<<"\t"<<avg_ll<<"\t"<<avg_l<<endl;
	llfile<<"STD"<<"\t"<<std_ll<<"\t"<<std_l<<endl;
	llfile.close();
}





void traitRateMng::simulateNullModel()
{
	string baseOutDirName = traitRateOptions::_outDir;
	traitRateOptions::_relRate = 1.0;
    
	//the model is the same for all tree iterations and doesn't change due to optimizations
	_pSeqAlph = new nucleotide;
	getTree();
	initStartingModel();

	_pSeqAlph = new nucleotide;
    for (int trIt = traitRateOptions::_startSimulationsIter; trIt <= traitRateOptions::_endSimulationsIter; ++trIt)
	{
		//open simulated directory
		string iterOutDirName = baseOutDirName + "//sim_" + int2string(trIt);
        traitRateOptions::_outDir = iterOutDirName;
		createDir("", iterOutDirName);
		
        printOptionParameters();
        simulateCombinedModel sim(_tree, _pCombinedModel, &_charAlph, _pSeqAlph);
		sim.simulate(traitRateOptions::_simulationSeqPos);
		sequenceContainer simCharContainer = sim.toCharDataWithoutInternalNodes();
		sequenceContainer simSeqContainer = sim.toSeqDataWithoutInternalNodes();
		string outCharName = iterOutDirName + "//simChars.fasta";
		string outSeqName = iterOutDirName + "//simSeqs.fasta";
		string outPhylip = iterOutDirName + "//simSeqs.phy";
		string outTreeName = iterOutDirName + "//model.tree";
		ofstream outChar(outCharName.c_str());
		ofstream outSeq(outSeqName.c_str());
		ofstream outPhy(outPhylip.c_str());
		fastaFormat::write(outChar, simCharContainer);
		fastaFormat::write(outSeq, simSeqContainer);
		phylipFormat::write(outPhy, simSeqContainer);
		_tree.output(outTreeName);
		outChar.close();
		outSeq.close();
		outPhy.close();
	}
}

//(1) simulate a trait that cause rate shift, 
//(2) simulate sequences based on this trate, 
//(3) simulate another trait that will be used as the "false positive" trait
void traitRateMng::simulateFalsePositive()
{
	string baseOutDirName = traitRateOptions::_outDir;
	    
	//the model is the same for all tree iterations and doesn't change due to optimizations
	_pSeqAlph = new nucleotide;
	getTree();
	initStartingModel();

	if (traitRateOptions::_characterFile == "") {
		errorMsg::reportError("Error: missing character data file in the parameters file - simulations are based on existing character data");
	}
	ifstream inCharFile(traitRateOptions::_characterFile.c_str());
	_charContainer = recognizeFormat::read(inCharFile, &_charAlph);
	inCharFile.close();

	for (int trIt = traitRateOptions::_startSimulationsIter; trIt <= traitRateOptions::_endSimulationsIter; ++trIt)
	{
		//open simulated directory
		string iterOutDirName = baseOutDirName + "//sim_" + int2string(trIt);
        traitRateOptions::_outDir = iterOutDirName;
		createDir("", iterOutDirName);
		
        printOptionParameters();
		//simulate real trait-shift
        simulateCombinedModel sim(_tree, _pCombinedModel, &_charAlph, _pSeqAlph);
		sim.simulate(traitRateOptions::_simulationSeqPos, &_charContainer, 0.9);
		sequenceContainer simCharContainer1 = sim.toCharDataWithoutInternalNodes();
		sequenceContainer simSeqContainer = sim.toSeqDataWithoutInternalNodes();
		string outCharName1 = iterOutDirName + "//simRandomChars.fasta";
		string outTreeName = iterOutDirName + "//model.tree";
		ofstream outChar1(outCharName1.c_str());
		fastaFormat::write(outChar1, simCharContainer1);
		_tree.output(outTreeName);
		outChar1.close();
	}
}


void traitRateMng::simulateRRFixedTaxa()
{
	string baseOutDirName = traitRateOptions::_outDir;
	int rIter = 9;
	initAlphabet();
	for (int i = 0; i < rIter; ++i)
	{
		//MDOUBLE rr = pow(2.0, i);
		int mul = static_cast<int>(pow(2.0, i-1));
		MDOUBLE rr = 1.0 + 0.1*mul;
		//int rr = static_cast<int>(pow(2.0, i));
		string rrIterOutDirName = baseOutDirName + "//r" + double2string(rr, 2);
		string paramsFileName =rrIterOutDirName+ "//modelParams.txt";

		createDir("", rrIterOutDirName);
		
		//the model is the same for all tree iterations and doesn't change due to optimizations
		traitRateOptions::_relRate = rr;
        initStartingModel();
		ofstream paramFile(paramsFileName.c_str());
		_pCombinedModel->printModelParams(paramFile);
		paramFile.close();


		for (int trIt = traitRateOptions::_startSimulationsIter; trIt <= traitRateOptions::_endSimulationsIter; ++trIt)
		{
			//open simulated directory
			string iterOutDirName = rrIterOutDirName + "//tree" + int2string(trIt);

			traitRateOptions::_outDir = iterOutDirName;
			createDir("", iterOutDirName);
			getSimulatedTree(trIt);
            printOptionParameters();

			simulateCombinedModel sim(_tree, _pCombinedModel, &_charAlph, _pSeqAlph);
			sim.simulate(traitRateOptions::_simulationSeqPos);
			sequenceContainer simCharContainer = sim.toCharDataWithoutInternalNodes();
			sequenceContainer simSeqContainer = sim.toSeqDataWithoutInternalNodes();
			string outCharName = iterOutDirName + "//simChars.fasta";
			string outSeqName = iterOutDirName + "//simSeqs.fasta";
			string outPhylip = iterOutDirName + "//simSeqs.phy";
			string outTreeName = iterOutDirName + "//model.tree";
			ofstream outChar(outCharName.c_str());
			ofstream outSeq(outSeqName.c_str());
			ofstream outPhy(outPhylip.c_str());
			fastaFormat::write(outChar, simCharContainer);
			fastaFormat::write(outSeq, simSeqContainer);
			phylipFormat::write(outPhy, simSeqContainer);
			_tree.output(outTreeName);
			outChar.close();
			outSeq.close();
			outPhy.close();
		}
	}
}

//
void traitRateMng::printOptionParameters()
{
	LOGDO(5, Parameters::dump(myLog::LogFile())); 
	LOG(traitRateOptions::_logValue, <<"\n ---------------------- THE PARAMETERS ----------------------------"<<endl;);
	if (traitRateOptions::_mainType.size() > 0) 
		LOG(traitRateOptions::_logValue, <<"main type: "<<traitRateOptions::_mainType<<endl;);
	if (traitRateOptions::_treeFile.size() > 0) 
		LOG(traitRateOptions::_logValue, <<"tree file: "<<traitRateOptions::_treeFile<<endl;);
	if (traitRateOptions::_treesFile.size() > 0) 
		LOG(traitRateOptions::_logValue, <<"trees file: "<<traitRateOptions::_treesFile<<endl;);
	if (traitRateOptions::_seqFile.size() > 0) 
		LOG(traitRateOptions::_logValue, <<"sequence file: "<<traitRateOptions::_seqFile<<endl;);
	if (traitRateOptions::_characterFile.size() > 0) 
		LOG(traitRateOptions::_logValue, <<"characters file: "<<traitRateOptions::_characterFile<<endl;);
	if (traitRateOptions::_outFile.size() > 0) 
		LOG(traitRateOptions::_logValue, <<"output file: "<<traitRateOptions::_outFile<<endl;);
	
	if 	(traitRateOptions::_rootAt.size() > 0) 
		LOG(traitRateOptions::_logValue, <<"root sequence name: "<<traitRateOptions::_rootAt<<endl;);
	LOG(traitRateOptions::_logValue, <<"\n -----------------------------------------------------------------"<<endl;);
}


void traitRateMng::getStartingData()
{
	if (traitRateOptions::_characterFile == "") {
		errorMsg::reportError("Error: missing character data file in the parameters file");
	}
	if (traitRateOptions::_seqFile == "") {
		errorMsg::reportError("Error: missing aligned sequence data file in the parameters file");
	}
	ifstream inCharFile(traitRateOptions::_characterFile.c_str());
	_charContainer = recognizeFormat::read(inCharFile, &_charAlph);
	inCharFile.close();
	initAlphabet();
	ifstream inSeqFile(traitRateOptions::_seqFile.c_str());
	_seqContainer = recognizeFormat::read(inSeqFile, _pSeqAlph);
    inSeqFile.close();
	_seqContainer.changeGaps2MissingData();

	//check that char container includes only one character
	Vint type4Pos;
	getCharacterType4pos(_charContainer, type4Pos);
	if (type4Pos[0] < 2)
		errorMsg::reportError("only one character type found in the character data");
	
}

void traitRateMng::getTree() {
	if (!traitRateOptions::_treesFile.empty())
		return getTrees();
	if (traitRateOptions::_treeFile.empty())
		errorMsg::reportError("Error: missing tree file in the parameters file");		
    _tree = tree(traitRateOptions::_treeFile);
	if (!_tree.withBranchLength()) 
		errorMsg::reportError("Error: branch lengths should be specified in the input tree");		
	MDOUBLE treeLength = getSumOfBranchLengths(_tree);
	MDOUBLE reqTreeLen = traitRateOptions::_initTreeLength;
	if (!DEQUAL(0.0, reqTreeLen, 0.0001))
	{
		if (reqTreeLen < 0.0)
		{//in this case _initTreeLength indicates the average length from the root the the leaves
			vector<tree::nodeP> leaves;
			_tree.getAllLeaves(leaves, _tree.getRoot());
			//we assume tree is ultrametric
			MDOUBLE tip2root = _tree.findLengthBetweenAnyTwoNodes(leaves[0], _tree.getRoot());
			_tree.multipleAllBranchesByFactor(-reqTreeLen / tip2root);
		}
		else // the finalTreeLength should be _initTreeLength
            _tree.multipleAllBranchesByFactor(reqTreeLen / treeLength);
		treeLength = getSumOfBranchLengths(_tree);
	}

	if (!traitRateOptions::_referenceTree.empty())
	{
		tree refTree(traitRateOptions::_referenceTree);
		traitRateOptions::_treeLength = getSumOfBranchLengths(refTree);
		cerr<<"ref tree len = "<<traitRateOptions::_treeLength<<endl;
	}
	if (traitRateOptions::_treeLength > 0) 
	{
		LOGnOUT(10,<<"original tree length is "<<treeLength<<endl);
		_tree.multipleAllBranchesByFactor(traitRateOptions::_treeLength /treeLength);
		treeLength = getSumOfBranchLengths(_tree);
	}

	LOG(3, <<"total tree length = "<<treeLength<<endl);
	LOG(3, <<"required tree length = "<<traitRateOptions::_treeLength<<endl);
	//print tree with internal nodes
	string outFileName = traitRateOptions::_outDir + "//" + "allNodes.tree";
	_tree.output(outFileName, tree::PHYLIP, true);
	if (_treeVec.empty() && (_tree.getNodesNum() > 0))
		_treeVec.push_back(_tree);
}

void traitRateMng::getTrees() {
	if (!traitRateOptions::_treeFile.empty())
		errorMsg::reportError("Error: cannot run both multiple trees file and a single tree file");		

    vector<tree> allTrees = getStartingTreeVecFromFile(traitRateOptions::_treesFile);
	if (allTrees.empty())
		errorMsg::reportError("error reading tree list");

	if (allTrees.size() != traitRateOptions::_stochasicMappingIterations)
		_treeVec = traitRateOptimizer::pickTrees(allTrees, traitRateOptions::_stochasicMappingIterations);
	else
		_treeVec = allTrees;

	if (_treeVec.empty())
		errorMsg::reportError("Error: empty trees file");		

	if (!traitRateOptions::_referenceTree.empty())
	{
		tree refTree(traitRateOptions::_referenceTree);
		traitRateOptions::_treeLength = getSumOfBranchLengths(refTree);
		cerr<<"ref tree len = "<<traitRateOptions::_treeLength<<endl;
	}
	//if (traitRateOptions::_treeLength > 0) 
	//{
	//	for (int t = 0; t < _treeVec.size(); ++t) 
	//	{
	//		MDOUBLE treeLength = getSumOfBranchLengths(_treeVec[t]);
 //           _treeVec[t].multipleAllBranchesByFactor(traitRateOptions::_treeLength /treeLength);
	//	}
	//}

	
	//validate tree

	checkTreeAndContainerSameTaxa(_treeVec[0], _seqContainer, traitRateOptions::_seqFile, traitRateOptions::_treesFile);
	checkTreeAndContainerSameTaxa(_treeVec[0], _charContainer, traitRateOptions::_characterFile, traitRateOptions::_treesFile);
	_treeVec[0].isUltrametric(0.05, true);

	//if tree is not rooted then user must speicify the rooting position
	if((_treeVec[0].getRoot()->getNumberOfSons() > 2) && (traitRateOptions::_rootAt == ""))
	{
		if (traitRateOptions::_rootFreqType != twoStateModel::STATIONARY)
            errorMsg::reportError("The input tree is not rooted. The root frequencies must be estimated according to stationary distribution");
	}
}


//1. check that all taxa in the tree are present in both _charContainer and _seqContainer
//2. check that tree is ultrametric
//3. make sure tree is rooted in case the root frequencies are not estiated according to stationarity of the model
void traitRateMng::validateInputs()
{
	if (!traitRateOptions::_treeFile.empty()) 
	{
		checkTreeAndContainerSameTaxa(_tree, _seqContainer, traitRateOptions::_seqFile, traitRateOptions::_treeFile);
		checkTreeAndContainerSameTaxa(_tree, _charContainer, traitRateOptions::_characterFile, traitRateOptions::_treeFile);
		_tree.isUltrametric(0.05, true);

		//if tree is not rooted then user must speicify the rooting position
		if((_tree.getRoot()->getNumberOfSons() > 2) && (traitRateOptions::_rootAt == ""))
		{
			if (traitRateOptions::_rootFreqType != twoStateModel::STATIONARY)
				errorMsg::reportError("The input tree is not rooted. The root frequencies must be estimated according to stationary distribution");
		}
	}
	//validate parameters of NULL model
	if (traitRateOptions::_modelType == M0)
	{
		if (traitRateOptions::_bOptCharModel || traitRateOptions::_bOptRelRate)
			errorMsg::reportError("cannot optimize character model parameters using model M0");
	}

	if ((traitRateOptions::_bScaleTree == 0) && (traitRateOptions::_treeLength <= 0) && traitRateOptions::_referenceTree.empty())
		errorMsg::reportError("A reference tree is missing. The option _bScaleTree should be set to 1");

	if ((traitRateOptions::_sequenceType == PROTEIN) && (_pSeqAlph->size()!=20))
		errorMsg::reportError("mismatch between the sequence type and the sequence evolutionary model");
	if ((traitRateOptions::_sequenceType == NON_CODING) && (_pSeqAlph->size()!=4))
		errorMsg::reportError("mismatch between the sequence type and the sequence evolutionary model");
}

//init both the character process and sequence process
void traitRateMng::initStartingModel()
{
	if (_pCombinedModel)
		delete _pCombinedModel;

	//init character model
	twoStateModel charModel(traitRateOptions::_charModelParam1, traitRateOptions::_charModelParam2, getInitialCharFreq(), traitRateOptions::_rootFreqType);
	trivialAccelerator pijAccChar(&charModel);
    uniDistribution uniDistr;
	stochasticProcess* pCharSp = new stochasticProcess(&uniDistr, &pijAccChar, isCharModelReversible());
	
	//init sequence model
	replacementModel *pProbMod=NULL;
	pijAccelerator *pPijAccSeq=NULL;
	Vdouble initFreqs(_pSeqAlph->size(), 1.0 / _pSeqAlph->size());
	switch (traitRateOptions::_seqModelType){
		case JC:
			pProbMod = new nucJC;
			pPijAccSeq = new trivialAccelerator (pProbMod);
			break;
		case K2:
			pProbMod = new hky(initFreqs, traitRateOptions::_seqModelParam1); 
			pPijAccSeq = new trivialAccelerator (pProbMod);
			break;
		case HKY:
			if (_seqContainer.numberOfSeqs() > 0)
                initFreqs = evaluateCharacterFreq(_seqContainer);
			pProbMod = new hky(initFreqs, traitRateOptions::_seqModelParam1); 
			pPijAccSeq = new trivialAccelerator (pProbMod);
			break;
		case GTR:
			if (_seqContainer.numberOfSeqs() > 0)
                initFreqs = evaluateCharacterFreq(_seqContainer);
			pProbMod = new gtrModel(initFreqs, traitRateOptions::_seqModelParam1, traitRateOptions::_seqModelParam2, traitRateOptions::_seqModelParam3, traitRateOptions::_seqModelParam4, traitRateOptions::_seqModelParam5, traitRateOptions::_seqModelParam6); 
			pPijAccSeq = new trivialAccelerator (pProbMod);
			break;
		case JTT:
			pProbMod = new pupAll(datMatrixHolder::jones); 
			pPijAccSeq = new chebyshevAccelerator(pProbMod); 
			break;
		case MT_REV:
			pProbMod = new pupAll(datMatrixHolder::mtREV24); 
			pPijAccSeq = new chebyshevAccelerator(pProbMod); 
			break;
		case CP_REV:
			pProbMod = new pupAll(datMatrixHolder::cpREV45); 
			pPijAccSeq = new chebyshevAccelerator(pProbMod); 
			break;
		case LG:
			pProbMod = new pupAll(datMatrixHolder::lg); 
			pPijAccSeq = new chebyshevAccelerator(pProbMod); 
			break;
		case WAG:
			pProbMod = new pupAll(datMatrixHolder::wag); 
			pPijAccSeq = new chebyshevAccelerator(pProbMod); 
			break;
		case AA_JC:
			pProbMod = new aaJC; 
			pPijAccSeq = new trivialAccelerator(pProbMod); 
			break;
		default:
			errorMsg::reportError("unknown sequence model in traitRateMng::initStartingModel");
	}
	
	distribution *pDist = NULL;
	if (traitRateOptions::_gammaCategories == 1)
		pDist = new uniDistribution;
	else if (traitRateOptions::_gammaCategories > 1)
        pDist = new gammaDistribution(traitRateOptions::_gammaParam, traitRateOptions::_gammaCategories);
	else
		errorMsg::reportError("Number of gamma categories must be positive");

	pDist->setGlobalRate(traitRateOptions::_globalRate);
	stochasticProcess* pSeqSp = new stochasticProcess(pDist, pPijAccSeq);
	_pCombinedModel = new charAndSeqModel(pCharSp, pSeqSp, traitRateOptions::_relRate, traitRateOptions::_seqModelType, traitRateOptions::_modelType, traitRateOptions::_treeLength);
	if (_charContainer.seqLen() > 0)
        _pCombinedModel->setParam1Bound(_charContainer);
	//calculate stats to determine upper/lower bound for characterModel param 2
	//if (traitRateOptions::_mainType == "simulateRRFixedTaxa")
	if (traitRateOptions::_mainType.find("simulate")  != -1)
	{
		_pCombinedModel->setParam2Bound(10);
	}
	else {
		MDOUBLE sumLength = 0.0;
		for (int t = 0; t < _treeVec.size(); ++t)
			sumLength += getSumOfBranchLengths(_treeVec[t]);
		MDOUBLE treeLength = sumLength / static_cast<MDOUBLE>(_treeVec.size());

		int mpScore = getMPScore(_treeVec[0], _charContainer, 0);
		MDOUBLE param2multiplier = mpScore / treeLength;
		_pCombinedModel->setParam2Bound(param2multiplier);
		_pCombinedModel->setTreeLengthBound(treeLength);
		if (traitRateOptions::_treeLength > 0) 
			treeLength = traitRateOptions::_treeLength;
		_pCombinedModel->setTreeLength(treeLength);
	}

	delete pSeqSp;
	delete pCharSp;
	delete pDist;
	delete pProbMod;
	delete pPijAccSeq;
}


//returns the frequency of character 1
MDOUBLE traitRateMng::getInitialCharFreq()
{
	MDOUBLE freq1 = -1;
	switch (traitRateOptions::_rootFreqType){
		case (twoStateModel::EMPIRICAL):
			{
			Vdouble freqs = evaluateCharacterFreq(_charContainer); 
			freq1 = freqs[1];
			LOGnOUT(4,<<"frequencies are based on LeavesAve "<<freqs[0]<<" "<<freqs[1]<<endl);
			}
			break;
		case twoStateModel::STATIONARY:
			return freq1;
		case twoStateModel::ROOT_LL:
		default:
			errorMsg::reportError("not handled yet");
	}

	return freq1;
}

bool traitRateMng::isCharModelReversible()
{
	if (traitRateOptions::_rootFreqType == twoStateModel::STATIONARY)
		return true;
	else
        return false;
}

//1. optimize the sequence stochastic process. 
//2. scale the tree to get the best likelihood.
void traitRateMng::optimizeTreeAndInitialSeqModel()
{
	LOGnOUT(5,<<"Optimizing Model M0"<<endl);
	combinedModelType modelType = _pCombinedModel->getCombinedModelType();
	_pCombinedModel->setCombinedModelType(M0);
	MDOUBLE relRate = _pCombinedModel->getRelativeRate();
	_pCombinedModel->setRelativeRate(1.0);
	int optIter = 2; //do only a small number of optimizations
	traitRateOptimizer opt(_seqContainer, _charContainer);
	vector<paramType> params = getOptimizedParams();

	if (traitRateOptions::_mainType == "Run_Fix_Param") {
		_model0LL = _pCombinedModel->compLogLikelihood(_treeVec, _seqContainer, _treeVec, _charContainer, traitRateOptions::_llCalcType);
	}
	else 
		_model0LL = opt.optimizeCombinedModel(params, _treeVec, _pCombinedModel, traitRateOptions::_epsilonLLimprovement, optIter, traitRateOptions::_tolParamOptimization, _treeVec.size(), 0, traitRateOptions::_llCalcType);
    ////debug
	//MDOUBLE treeLength = getSumOfBranchLengths(_tree);
	//MDOUBLE globalR = _pCombinedModel->getGlobalRate();
	//MDOUBLE reqTreeLen = treeLength * globalR;
	//if (_treeVec.size() == 1)
 //       _treeVec[0].multipleAllBranchesByFactor(reqTreeLen / treeLength);
	//else
	//	errorMsg::reportError("treeVec size should be one");
	//treeLength = getSumOfBranchLengths(_treeVec[0]);
  //  _pCombinedModel->setGlobalRate(1.0);
	//MDOUBLE model0LL = _pCombinedModel->compLogLikelihood(_treeVec, _seqContainer, _treeVec, _charContainer, traitRateOptions::_llCalcType);
	//cerr<<"LL before = "<<_model0LL<<endl;
	//cerr<<"LL after = "<<model0LL<<endl;

	//////
	_pCombinedModel->setCombinedModelType(modelType);
	_pCombinedModel->setRelativeRate(relRate);
}


MDOUBLE traitRateMng::optimizeCombinedModel()
{
	LOGnOUT(5,<<"Optimizing Model M1"<<endl);
    MDOUBLE res = -1;
	if (traitRateOptions::_mainType == "Run_Fix_Param")
	{
		if (traitRateOptions::_modelType == M0)
			res = _pCombinedModel->compLogLikelihood(_treeVec, _seqContainer, _treeVec, _charContainer, traitRateOptions::_llCalcType);
		else
		{
			_treeVec = traitRateOptimizer::pickTrees(_treeVec, traitRateOptions::_stochasicMappingIterations);
			res= _pCombinedModel->compLogLikelihood(_treeVec, _charContainer, _seqContainer, traitRateOptions::_stochasicMappingIterations, 0, _treeVec, traitRateOptions::_llCalcType); 

		}
	}
	else if (traitRateOptions::_mainType == "Optimize_Model")
	{
		traitRateOptimizer opt(_seqContainer, _charContainer);
		vector<paramType> optParams = getOptimizedParams();

		if (traitRateOptions::_modelType == M0)
		{
			res = opt.optimizeCombinedModel(optParams, _treeVec, _pCombinedModel, traitRateOptions::_epsilonLLimprovement, traitRateOptions::_maxOptimizationIterations, traitRateOptions::_tolParamOptimization, 0, 0, traitRateOptions::_llCalcType);
		}
		else 
		{
			time_t t1;
			time(&t1);
			LOGnOUT(5,<<"Optimizing parameters"<<endl);
			Vdouble tols(traitRateOptions::_optimizeIterNum.size(), traitRateOptions::_epsilonLLimprovement);
			res = opt.optimizeModelManyStart(optParams, _treeVec, _pCombinedModel, traitRateOptions::_optimizePointsNum, traitRateOptions::_optimizeIterNum, tols, traitRateOptions::_optimizeStartegies, traitRateOptions::_stochasicMappingIterations, traitRateOptions::_llCalcType);
			LOG(3, <<"after optmizations"<<endl;);
			int cycles = traitRateOptions::_optimizeStartegies.size();
			if (traitRateOptions::_optimizeStartegies[cycles-1] != 0) //make sure to compute exact LL in case LL computation in the last cycle was heuristic  
				res = _pCombinedModel->compLogLikelihood(_treeVec, _seqContainer, _charContainer, traitRateOptions::_stochasicMappingIterations, 0, _treeVec, traitRateOptions::_llCalcType);
			time_t t2;
			time(&t2);
			LOGnOUT(5,<<endl<<"TOTAL RUNNING TIME = "<<static_cast<int>(t2-t1)<<endl);
		}
	}
	return res;
}

const vector<paramType> traitRateMng::getOptimizedParams()
{
	vector<paramType> res;
	if (traitRateOptions::_bScaleTree)
        res.push_back(TREE_LENGTH);
	if (traitRateOptions::_bOptRelRate && (_pCombinedModel->getCombinedModelType() != M0))
		res.push_back(RELATIVE_RATE);
	if (traitRateOptions::_bOptCharModel)
		_pCombinedModel->getCharModelParams(res);
	if (traitRateOptions::_bOptSeqModel)
		_pCombinedModel->getSeqModelParams(res);
	return res;
}

void traitRateMng::printResults()
{
	string fileName = traitRateOptions::_outDir + "//" + traitRateOptions::_outFile;
	ofstream outF(fileName.c_str());
	outF<<"#Model: "<<traitRateOptions::getCombinedModelTypeStr(traitRateOptions::_modelType)<<endl;
	outF<<"#Input character data file: "<<traitRateOptions::_characterFile<<endl;
	outF<<"#Input sequence data file: "<<traitRateOptions::_seqFile<<endl;
	if (!traitRateOptions::_treeFile.empty())
        outF<<"#Input tree file: "<<traitRateOptions::_treeFile<<endl;
	if (!traitRateOptions::_treesFile.empty())
		outF<<"#Input tree file: "<<traitRateOptions::_treesFile<<endl;
	outF<<endl;
	outF<<"#total tree length = "<<getSumOfBranchLengths(_treeVec[0])<<endl;
	outF<<"#Character model root frequencies: "<<traitRateOptions::getRootFreqTypeStr(traitRateOptions::_rootFreqType)<<endl;
	MDOUBLE f0 = static_cast<twoStateModel*>(_pCombinedModel->getCharProcess()->getPijAccelerator()->getReplacementModel())->freq(0);
	MDOUBLE f1 = static_cast<twoStateModel*>(_pCombinedModel->getCharProcess()->getPijAccelerator()->getReplacementModel())->freq(1);
	outF<<"#F[0] = "<<f0<<" F[1] = "<<f1<<endl;

	//print params
	_pCombinedModel->printModelParams(outF);

	outF << "LogLikelihood = "<<_bestLL<<endl;
	outF << "LogLikelihood Model 0 = "<<_model0LL<<endl;
	outF << "Likelihood = "<<exp(_bestLL)<<endl;
	outF.close();

	string scaledTreeStr = traitRateOptions::_outDir + "//" + traitRateOptions::_scaledTreeFile;
	printScaledTree(scaledTreeStr);
}

void traitRateMng::printScaledTree(const string& scaledTreeStr)
{
	tree scaledTree = _pCombinedModel->getAvgScaledTree(_treeVec[0], _charContainer, traitRateOptions::_stochasicMappingIterations);

	//change names to leaves
	seqContainerTreeMap scTreeMap(_charContainer, scaledTree);	
	vector <tree::nodeP> leaves;
	scaledTree.getAllLeaves(leaves, scaledTree.getRoot());
	for (int i=0; i< leaves.size(); ++i){
		int myleafId = (leaves[i])->id();
		int mySeqId = scTreeMap.seqIdOfNodeI(myleafId);
		int state = _charContainer[mySeqId][0];
		string newName = leaves[i]->name() + "-" + _charAlph.fromInt(state);
		leaves[i]->setName(newName);
	}
	scaledTree.output(scaledTreeStr);
}


void traitRateMng::getSimulatedTree(int iter)
{
	if (traitRateOptions::_simulationsTreeDir.empty())
		return getTree();
	string treeFile = traitRateOptions::_simulationsTreeDir + "/" + int2string(iter) + ".tree";
	traitRateOptions::_treeFile = treeFile;
	getTree();
	_tree.isUltrametric(0.05, true);
}

void traitRateMng::initAlphabet()
{
	if (_pSeqAlph != NULL)
		delete _pSeqAlph;
	switch (traitRateOptions::_seqModelType)
	{
		case JC:
        case K2:
		case HKY:
		case GTR:
			_pSeqAlph = new nucleotide;
			break;
		case JTT:
		case MT_REV:
		case CP_REV:
		case LG:
		case WAG:
		case AA_JC:
			_pSeqAlph = new amino;
			break;
		default:
			errorMsg::reportError("unknown sequence model in traitRateMng::initAlphabet");
	}
}
