#ifndef ___2STATE_MODEL
#define ___2STATE_MODEL

/*
Q is a matrix of the following form:
	0		1		
0  [-pi1  pi1] x r	
1  [ pi0 -pi0]	
where pi1 = 1-pi0; 0<pi0< 1
The rate parameter r is used to transform the branch lengths of the tree, usually given in avegrage # of nucleotide or aa substitutions
to the units of the 2-state model.

//In fact the matrix can be written in its normalized form (so that the total amount of change )
//by multiplying the whole matrix by 1/2pi0pi1. 
//BUT - THE CURRENT IMPLEMENTATION DOESN'T NORMALIZE THE MATRIX BECAUSE IT DOESN'T MATTER IF WE HAVE THE _RQ PARAM. 
//	0		1		
//0  [-1/2pi0  1/2pi0] x r	
//1  [1/2pi1  -1/2pi1]	

We can also constrain the model such that q_01=q_10 so that there is a single free parameter, r.

ROOT FREQENCIES
Another component of the model is how to determine the frequencies at the root.
In case there is only one position (one trait) the root frequencies cannot be estimated from the data because it will always results in one of the state with the lower likelihood to have frequency = 0. 
Other possibilites:
1. the root frequencies can be set to the stationary distribution of the rate matrix.
In this case the process will be stationary (and reversible??)
ROOT_F[0] = q_10/q_10+q_01, ROOT_F[1] = q_01/q_10+q_01 = 1-ROOT_F[0]

2. Empirically: 
According to the two-character distribution of the inout data, without taking into account the phylogeny
3. According to the likelihood at the root: 
Let L(r=0) = P(D|root_state = 1) 
then L(D) = L(r=0) + L(r=1)
and ROOT_F[0] = L(r=0)/L(D), ROOT_F[1] = L(r=1)/L(D)
*/


#include "definitions.h"
#include "replacementModel.h"
#include "errorMsg.h"

class twoStateModel : public replacementModel {
public:
	//ROOT_LL: the frequencies are according to the likelihood at the root: f(0) = L(root=0)/{L(root=0)+L(root=1)}
	//EMPIRICAL: according to the character distribution at the leaves
    enum rootFreqType {ROOT_LL, STATIONARY, EMPIRICAL}; 
public:
	explicit twoStateModel(MDOUBLE pi1, MDOUBLE globalRate, MDOUBLE freq1, rootFreqType rootType);
	twoStateModel(const twoStateModel& other) {*this = other;}
	virtual twoStateModel& operator=(const twoStateModel &other);
	virtual twoStateModel* clone() const { return new twoStateModel(*this); }
	virtual ~twoStateModel() {}
	const int alphabetSize() const {return 2;} 
	const MDOUBLE err_allow_for_pijt_function() const {return 1e-4;} // same as q2p definitions
	const MDOUBLE Pij_t(const int i,const int j, const MDOUBLE d) const ;		
	const MDOUBLE dPij_dt(const int i,const int j, const MDOUBLE d) const; 
	const MDOUBLE d2Pij_dt2(const int i,const int j, const MDOUBLE d) const;
	const MDOUBLE freq(const int i) const;

	const Vdouble getFreqs() const  {return _freq;}
	void setFreq(MDOUBLE f1); 
	MDOUBLE getPi1() const {return _pi1;}
	void setPi1(const MDOUBLE val);
	MDOUBLE getScaleRate() const {return _rQ;};
	void setScaleRate(const MDOUBLE val);
	rootFreqType getRootType() const {return _rootType;}
	
private:
	virtual void updateQ();
	bool pijt_is_prob_value(MDOUBLE val) const;
	bool areFreqsValid(Vdouble freq) const; // tests if frequencies are valid (>0, sum=1)

private:
	VVdouble _Q;
	Vdouble _freq;
	MDOUBLE _rQ;
	MDOUBLE _pi1;
	rootFreqType _rootType;
};
#endif // ___2STATE_MODEL
